// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.logic;


import java.util.HashMap;
import java.util.Map;

import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.inst.SVInstantiations;


/** Meta operator for inlining named ASM rules. This meta operator
 * replaces a named asm rule invocation with the definition of the
 * named asm rule and substitutes the formal parameters with the
 * arguments of the rule invocation.
 *
 * @author Hubert Schmid
 */

public final class MetaNamedRule extends AsmMetaOperator {

    /** constructs a new meta operator. Only one instance is required,
     * so this may be a singleton class. */
    public MetaNamedRule() {
        super("ASM_NAMED_RULE", AsmSort.ASMRULE,
              new Sort[] { AsmSort.ASMRULE });
    }


    /** Returns the body of the named rule with parameter substitution
     * already performed. 'term' contains one subterm -- the named
     * rule invocation.  The named rule is determined by the operator
     * of this subterm. */
    public Term calculate(Term term, SVInstantiations svInst) {
        Term call = term.sub(0);
        AsmProcedure p = ((AsmProcedure.Operator) call.op()).getProcedure();
        FormalParameter[] params = p.getFormalParameters();
        if (call.arity() != params.length) {
            throw new RuntimeException("Procedure call und procedure have different arities.");
        }
        Map map = new HashMap();
        for (int i = 0; i < params.length; ++i) {
            map.put(params[i], call.sub(i));
        }
        return FormalParameterSubstitution.apply(map, p.getBody());
    }

}
