// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.env;


import de.uka.ilkd.asmkey.logic.AsmProcedure;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.*;


/** This interface represents the environment (declarations) necessary
 * for parsing terms.
 *
 * @author Hubert Schmid
 */

public interface TermEnvironment extends SortEnvironment {

    /** Determines if a meta operator with the given name exists. */
    boolean containsMetaOp(String id);

    /** Returns the meta operator with the given name.
     *
     * @throws EnvironmentException if no meta operator with the given
     * name exists.
     */
    MetaOperator getMetaOp(String id) throws EnvironmentException;

    /** Determines if a sort with the given name exists. */
    boolean containsSchemaVariable(String id);

    /** Returns the schema variable with the given name.
     *
     * @throws EnvironmentException if no such schema variable exists.
     */
    SchemaVariable getSchemaVariable(String id) throws EnvironmentException;

    /** Returns the operator with the given name.
     *
     * @throws EnvironmentException if no such operator exists.
     */
    Operator getOperator(String id) throws EnvironmentException;

    /** Returns the procedure (named ASM ruls) with the given name.
     *
     * @throws EnvironmentException if no such procedure exists.
     */
    AsmProcedure getProcedure(String id) throws EnvironmentException;

    /** Returns the term which is abbreviated with name.
     *
     * @throws EnvironmentException if no such abbreviation exists.
     */
    Term getAbbreviatedTerm(String name) throws EnvironmentException;

}
