// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.tree;


import de.uka.ilkd.key.parser.ParserException;


/** This class wraps a ParserException into a parser exception. The
 * ParserException have to be declared in the throws clause. Some
 * methods don't allow ParserExceptions to be thrown (e.g. the
 * callbacks in the visitors). Hence the ParserException is wrapped
 * into a RuntimeException, thrown and then unwrapped.
 *
 * @author Hubert Schmid */

final class ParserExceptionWrapper extends RuntimeException {

    /** The wrapped exception. */
    private final ParserException cause;


    ParserExceptionWrapper(ParserException cause) {
        super(cause.getMessage());
        this.cause = cause;
    }


    /** Return the wrapped exception. */
    public ParserException getParserException() {
        return cause;
    }

}
