// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.tree;


import de.uka.ilkd.asmkey.logic.AsmSort;
import de.uka.ilkd.asmkey.parser.ast.*;
import de.uka.ilkd.asmkey.parser.env.DeclarationEnvironment;
import de.uka.ilkd.asmkey.parser.env.EnvironmentException;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.parser.ParserException;


/** Class to parse declarations of schema variables.
 *
 * @author Hubert Schmid
 */

final class SchemaVariableDeclarationParser {

    /** Parse the schema variable with the given identifier and the
     * given schema type, and add it to the environment 'env.' */
    protected static void parse(final DeclarationEnvironment env,
                                Identifier id, final AstSchemaType type) {
        final Name name = new Name(id.getText());

        type.accept(new AstSchemaTypeVisitor() {

                public void visitPrimitive(PrimitiveSchemaType t) {
                    try {
                        try {
                            if (t == PrimitiveSchemaType.FORMULA) {
                                env.addSchemaVariable(SchemaVariableFactory.createFormulaSV(name, false));
                            } else if (t == PrimitiveSchemaType.ASMRULE) {
                                env.addSchemaVariable(SchemaVariableFactory.createTermSV(name, AsmSort.ASMRULE, false));
                            } else {
                                throw new ParserException("unknown primitive schema type",
                                                          type.getLocation());
                            }
                        } catch (EnvironmentException ee) {
                            throw new ParserException(ee.getMessage(), type.getLocation());
                        }
                    } catch (ParserException pe) {
                        throw new ParserExceptionWrapper(pe);
                    }
                }

                public void visitComposed(Identifier sortId, boolean variable) {
                    try {
                        Sort sort = env.getSort(sortId.getText());
                        if (variable) {
                            env.addSchemaVariable(SchemaVariableFactory.createVariableSV(name, sort, false));
                        } else {
                            env.addSchemaVariable(SchemaVariableFactory.createTermSV(name, sort, false));
                        }
                    } catch (EnvironmentException ee) {
                        throw new ParserExceptionWrapper(new ParserException(ee.getMessage(),
                                                                             sortId.getLocation()));
                    }
                }

                public void visitDepending(Identifier sortId) {
                    try {
                        Sort sort = env.getSort(sortId.getText());
                        env.addSchemaVariable(SchemaVariableFactory.createDependingSV(name, sort, false));
                    } catch (EnvironmentException ee) {
                        throw new ParserExceptionWrapper(new ParserException(ee.getMessage(),
                                                                             sortId.getLocation()));
                    }
                }

            });
    }

}
