// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.rule;

import java.util.Collection;
import java.util.Iterator;
import de.uka.ilkd.asmkey.logic.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.BuiltInRule;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.util.Debug;


/** This class is a builtin rule to make a step from a world into the
 * (deterministic) subsequent world of the Kripke structure.
 *
 * @author Stanislas Nanchen
 */

public final class FunctionOperatorRule extends OperatorRule {
    
    /** Short refernce to TermFactory. */
    private static final TermFactory tf = TermFactory.DEFAULT;

    private final Name name = new Name("FunctionOperator");

    public boolean isApplicable(Goal goal, PosInOccurrence pos, Constraint constraint) {
	if (pos == null || pos.posInTerm() != PosInTerm.TOP_LEVEL) {
            return false;
        } else {
            Term term = pos.constrainedFormula().formula();
            if (term.op() instanceof FunctionOperator) {
		return true;
	    } else {
		return false;
	    }
        }
    }


     /** This method replaces a big operator by a suitable formula */
    protected Term calculate(Term term, Services services) {
	FunctionOperator function;
	Term formula;
	Term result;

	Debug.assertTrue(term.op() instanceof FunctionOperator);

	if (term.op() instanceof FunctionOperator) {
	    function = (FunctionOperator) term.op();
	    formula = term.sub(1);
	    Function f;

	    f = (Function) term.sub(0).op();
		

	    /* We create first n(arity) new logical variable and terms
	     * corresponding to them; and the f(x1, ..., xn) term
	     */
	    LogicVariable[] xs = new LogicVariable[f.arity()];
	    Term[] terms = new Term[f.arity()];
	    Term fterm;
	    VariableFactory factory = new VariableFactory();
	    for(int j = 0; j < xs.length; j++) {
		xs[j] = factory.createNewVariable(term, "x", f.argSort(j));
		terms[j] = tf.createVariableTerm(xs[j]);
	    }
	    fterm = tf.createFunctionTerm(f, terms);
	    
	    /* then we need to transform the formula of the big operator
	     * i.e. transform recursively the %f into:
	     * - a series of all quantors or ex quantors.
	     * - the fterm.
	     */
	    return transform_and_subst(formula, function.getVariableName(), xs, fterm, services);
	}
	return term;
    }
    

    public Name name() { return name; }

    public String displayName() {
        return "FunctionOperator";
    }

}
