// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.casetool;


import tudresden.ocl.check.types.*;
import tudresden.ocl.check.OclTypeException;

public class UMLOCLMultiplicity {

    public static int INFINITE = -10;

    public static UMLOCLMultiplicity STANDARD     = new UMLOCLMultiplicity(1,1);
    public static UMLOCLMultiplicity ZERO_TO_INF  = new UMLOCLMultiplicity(0,INFINITE);
    public static UMLOCLMultiplicity ONE_TO_INF   = new UMLOCLMultiplicity(1,INFINITE);
    public static UMLOCLMultiplicity ZERO_OR_ONE  = new UMLOCLMultiplicity(0,1);


    protected int min, max;

    public UMLOCLMultiplicity(int min, int max) {
	if ( min<0 )
	    throw new IllegalArgumentException("minimum multiplicity must be at least 0, is "+min);
	if ( max==0 )
	    throw new IllegalArgumentException("maximum multiplicity must not be 0");
	if ( !(min<=max  || max==INFINITE))
	    throw new IllegalArgumentException(
					       "minimum multiplicity ("+min+") must not be greater than maximum multiplicity ("+max+")"
					       );
	this.min=min;
	this.max=max;
    }

    public int getMax() {
	return max;
    }

    public int getMin() {
	return min;
    }

    public static UMLOCLMultiplicity getMultiplicityFromString(String s) {
	if (s.equals("0..*") || 
	    s.equals("*")){
	    return UMLOCLMultiplicity.ZERO_TO_INF;
	} else if (s.equals("1..*")){
	    return UMLOCLMultiplicity.ONE_TO_INF;
	} else if (s.equals("1")){
	    return UMLOCLMultiplicity.STANDARD;
	} else if (s.equals("0..1")){
	    return UMLOCLMultiplicity.ZERO_OR_ONE;
	}
	return null;
    }

    public String toString() {
	if (max==-10)	    
	    return ""+min+"..*";
	else
	    return ""+min+".."+max;
    }

}

