// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import java.io.*;
import java.awt.Point;
import javax.swing.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.pp.PosInSequent;

class QuantifierInstDropTargetListener implements DropTargetListener {
    
    private static final String ALL_LEFT = "all_left";
    private static final String EX_RIGHT = "ex_right";
    
    private SequentView seqView;
	
    QuantifierInstDropTargetListener(SequentView seqView) {
	this.seqView = seqView;
    }

    public void dragEnter (DropTargetDragEvent event) {}
		    
    public void dragExit (DropTargetEvent event) {}
		    
    public void dragOver (DropTargetDragEvent event) {}
		    
    public void drop (DropTargetDropEvent event) {
	String droppedString;
			
	Point dropLocation = event.getLocation();
			
	try {
	    Transferable transferable = event.getTransferable();

	    // we accept only Strings      
	    if (transferable.isDataFlavorSupported(DataFlavor.stringFlavor)){
		    
		event.acceptDrop(DnDConstants.ACTION_MOVE);
		droppedString = (String)transferable.
		    getTransferData(DataFlavor.stringFlavor);

		if(droppedString != null){

		    PosInSequent posInSeq = seqView
			.getPosInSequent(dropLocation);

		    //check if drop target is valid
		    if (posInSeq == null) {				    
			event.getDropTargetContext().dropComplete(true);
			return;
		    }
		    // get formula where String was dropped
		    ConstrainedFormula formula = 
			posInSeq.getPosInOccurrence().constrainedFormula();
					
		    // drop position must be the top level of a quantifier term
		    if (formula == null                                  ||
			!(formula.formula().op() instanceof Quantifier)  ||
			!posInSeq.getPosInOccurrence().isTopLevel()) {
			event.getDropTargetContext().dropComplete(true);
			return;
		    }

		    Operator quantifier = formula.formula().op();
		    // find right rule app depending on kind of quantifier
		    // and whether quantifier is in antecedent or succedent 
		    boolean droppenInAntec = 
			posInSeq.getPosInOccurrence().isInAntec();
		    SetOfTacletApp tacs = null;
		    if (quantifier == Op.ALL &&
			droppenInAntec)
			tacs = seqView.mediator().getTacletApplications
			    (seqView.mediator().getSelectedGoal(), ALL_LEFT,
			     posInSeq.getPosInOccurrence());
		    else if (quantifier == Op.EX &&
			     !droppenInAntec)
			tacs = seqView.mediator().getTacletApplications
			    (seqView.mediator().getSelectedGoal(), EX_RIGHT,
			     posInSeq.getPosInOccurrence());
			    
		    if (tacs == null) {
			event.getDropTargetContext().dropComplete(true);
			return;
		    }
		    if (tacs.size() != 1)
			throw new RuntimeException
			    ("Exception during DnD-instantiation "+
			     "occurred because there was not exactly "+
			     "one taclet application!");
		    IteratorOfTacletApp appIterator = 
			tacs.iterator();
		    TacletApp app = null;
		    while (appIterator.hasNext()) {
			app = appIterator.next();
		    }
			    
		    SetOfSchemaVariable uninstantiatedVars = 
			app.uninstantiatedVars();
		    if (uninstantiatedVars.size() != 1)
			throw new RuntimeException
			    ("Exception during DnD-instantiation "+
			     "occurred because there was not exactly "+
			     "one uninstantiated schema variable!");
		    IteratorOfSchemaVariable svIterator = 
			uninstantiatedVars.iterator();
		    while (svIterator.hasNext()) {
			SchemaVariable sv = svIterator.next();
			
			// put goal specific variables into namespace
			IteratorOfProgramVariable varIt = 
			    seqView.mediator().getSelectedGoal().
			    getGlobalProgVars().iterator();
			ProgramVariable pv;
			Namespace varNS = 
			    new Namespace(seqView.mediator().var_ns());
			while (varIt.hasNext()) {
			    pv = varIt.next();
			    varNS.add(pv);
			}
			NamespaceSet nss = 
			    seqView.mediator().namespaces().copy();
			nss.setVariables(varNS);

			TacletInstantiationsTableModel model = 
			    new TacletInstantiationsTableModel
			    (app, 
			     seqView.mediator().getServices(),
			     nss,
			     seqView.mediator().getNotationInfo().getAbbrevMap());
				
			int svRow = model.getSVRow(sv);
			model.setValueAt(droppedString, svRow, 1);
			try {
			    app = model.createTacletAppFromVarInsts();
			    seqView.mediator ().applyInteractive ( app,
                                    seqView.mediator ().getSelectedGoal () );
			}
			catch (SVInstantiationException exc) {
			    JOptionPane.showMessageDialog
				(new JFrame(), 
				 "Quantifier instantiation failed!\n"+exc,
				 "Error",  
				 JOptionPane.ERROR_MESSAGE);
			}
				
		    }
			    			
				    
		}
		event.getDropTargetContext().dropComplete(true);
	    } 
	    else{
		event.rejectDrop();
	    }
	}
	catch (IOException exception) {
	    exception.printStackTrace();
	    event.rejectDrop();
	} 
	catch (UnsupportedFlavorException ufException ) {
	    ufException.printStackTrace();						   
	    event.rejectDrop();
	}
    }
		    
    public void dropActionChanged(DropTargetDragEvent dtde) {
    }
}

