// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java;

import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 *  Package specification.
 * taken from COMPOST and changed to achieve an immutable structure
 */

public class PackageSpecification
 extends JavaNonTerminalProgramElement
 implements PackageReferenceContainer {


    /**
     *  Reference.
     */

    protected final PackageReference reference;

    /**
     * Package specification.
     * @param list an ExtList with children
     */

    public PackageSpecification(ExtList children) {
	super(children);
	reference=(PackageReference)children.get(PackageReference.class);
    }


    public SourceElement getLastElement() {
        return reference;
    }


    /**
 *      Returns the number of children of this node.
 *      @return an int giving the number of children of this node
    */

    public int getChildCount() {
        int result = 0;
        if (reference != null) result++;
        return result;
    }

    /**
 *      Returns the child at the specified index in this node's "virtual"
 *      child array
 *      @param index an index into this node's "virtual" child array
 *      @return the program element at the given position
 *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
 *                 of bounds
    */

    public ProgramElement getChildAt(int index) {
        int len;
        if (reference != null) {
            if (index == 0) return reference;
            index--;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0: reference
        if (child == reference) {
            return 0;
        }
        return -1;
    }


    /**
 *      Get package reference.
 *      @return the package reference.
     */

    public PackageReference getPackageReference() {
        return reference;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnPackageSpecification(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printPackageSpecification(this);
    }
}
