// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java.visitor;

import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.java.SourceElement;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.java.reference.VariableReference;
import java.util.HashSet;
import de.uka.ilkd.key.java.ProgramElement;

/** 
 * Walks through a java AST in depth-left-fist-order. 
 * This walker is used collect all ProgramVariables in a program.
 */
public class ProgramVariableCollector extends JavaASTVisitor {

    private HashSet result = new HashSet();

    public ProgramVariableCollector(ProgramElement root) {
	super(root);
    }

    /** the action that is performed just before leaving the node the
     * last time 
     */
    protected void doAction(ProgramElement node) {
	node.visit(this);
    }
    
    /** starts the walker*/
    public void start() {	
	walk(root());	
    }

    public HashSet result() { 
	return result;
    }    

    public String toString() {
	return result.toString();
    }

    protected void doDefaultAction(SourceElement x) {
    }

    public void performActionOnProgramVariable(ProgramVariable pv) {
	result.add(pv);
    }
    
    public void performActionOnMethodBodyStatement(MethodBodyStatement mbs) {
	if (mbs.getDesignatedContext() instanceof ProgramVariable) {
	    result.add(mbs.getDesignatedContext());
	} else if (mbs.getDesignatedContext() instanceof VariableReference) {
	    result.add(((VariableReference)mbs.getDesignatedContext()).getProgramVariable());
	}
    }

}
