// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.jml;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.declaration.modifier.Public;
import de.uka.ilkd.key.java.expression.literal.BooleanLiteral;
import de.uka.ilkd.key.java.expression.operator.CopyAssignment;
import de.uka.ilkd.key.java.expression.operator.New;
import de.uka.ilkd.key.java.recoderext.ClassInitializeMethodBuilder;
import de.uka.ilkd.key.java.recoderext.ImplicitFieldAdder;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.reference.TypeRef;
import de.uka.ilkd.key.java.reference.TypeReference;
import de.uka.ilkd.key.java.statement.CatchAllStatement;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ObjectSort;
import de.uka.ilkd.key.logic.sort.SortDefiningSymbols;
import de.uka.ilkd.key.proof.mgt.ListOfQuantifierPrefixEntry;
import de.uka.ilkd.key.proof.mgt.QuantifierPrefixEntry;

/** A generic jml methodspecification used for encapsulating lightweight
    and behavioral methodspecifications*/
public class JMLMethodSpec extends JMLSpec implements JMLLemmaMethodSpec, 
						      AssignableSpec{
    protected static int excCondCount = 0;
    protected static int resultVarCount = 0;
    protected static int logicVarCount = 0;
    protected static int exceptionVarCount = 0;
    // contains an assignable keyword (like nothing, everithing, ...), 
    // if one occurd in the specification.
    protected String assignableMode = null;
    protected ProgramVariable resultVar = null;
    protected TermFactory tf;
    protected MethodDeclaration md;
    //caches the classdeclaration from cSpec
    protected TypeDeclaration cd;
    protected JavaInfo ji;
    protected Services services;
    private Term pre;
    private Term post;
    protected HashMap lv2old;
    protected Term diverges;
    protected Namespace params;
    protected StatementBlock mBS = null;
    protected static Term trueLitTerm=null;
    protected static Term falseLitTerm=null;
    protected HashMap term2old;
    protected SetOfTerm assignableVariables = SetAsListOfTerm.EMPTY_SET;
    protected JMLClassSpec cSpec;
    protected SetOfSignals signals = SetAsListOfSignals.EMPTY_SET;
//    protected ListOfTerm signalCond = SLListOfTerm.EMPTY_LIST;
//    protected ListOfKeYJavaType signalExc = SLListOfKeYJavaType.EMPTY_LIST;
    protected static Term trueTerm = null;
    protected static Term falseTerm = null;
    protected static Term nullTerm = null;
    protected NamespaceSet nss;
    protected ProgramMethod pm;
    protected String inhFrom = "";
    protected ProgramVariable excVar;
    protected KeYJavaType exc;
    private Term pi1 = null;
    private ListOfQuantifierPrefixEntry oldLVs = null;
    private CatchAllStatement catchAllStmt;
    //contains the self-variable used in an inherited methodspec.
    protected ReferencePrefix inheritedPrefix;
    protected Term self; 


    public JMLMethodSpec(MethodDeclaration md, TermFactory tf, 
			 Services services, Namespace params, 
			 HashMap term2old, JMLClassSpec cSpec, 
			 NamespaceSet nss){
	this.md = md;
	this.cd = cSpec.getClassDeclaration();
	this.nss = nss;
	this.tf = tf;
	this.ji = services.getJavaInfo();
	this.services = services;
	pre = null;
	post = null;
	KeYJavaType returnType;
	if(md.getTypeReference() == null){
	    returnType = ji.getKeYJavaType("void");
	}else{
	    returnType = md.getTypeReference().getKeYJavaType();
	}
	pm = new ProgramMethod(md, ji.getKeYJavaType(cd), 
			       returnType, PositionInfo.UNDEFINED);
	if(nullTerm == null){
	    nullTerm = tf.createFunctionTerm(Op.NULL);
	}
	if(trueLitTerm == null){
	    trueLitTerm = services.getTypeConverter()
		.convertToLogicElement(BooleanLiteral.TRUE);
	}
	if(falseLitTerm == null){
	    falseLitTerm = services.getTypeConverter()
		.convertToLogicElement(BooleanLiteral.FALSE);
	}
	if(trueTerm == null){
	    trueTerm = tf.createJunctorTerm(Op.TRUE);
	}
	if(falseTerm == null){
	    falseTerm = tf.createJunctorTerm(Op.FALSE);
	}
	if(!md.isStatic()){
	    self = tf.createVariableTerm((ProgramVariable) 
					 cSpec.getInstancePrefix());
	    //adds self!=null && self.<created> == true or 
	    //self.<classInitialized> == true to the precondition
	    if(!(md instanceof Constructor)){
		Term t = tf.createEqualityTerm(self,nullTerm);
		addPre(tf.createJunctorTerm(Op.NOT, new Term[]{t}));
		ProgramVariable created = services.getJavaInfo().
		    getAttribute(ImplicitFieldAdder.IMPLICIT_CREATED, 
				 (ObjectSort) services.getJavaInfo().
				 getKeYJavaType(cd).getSort());
		Term eqc = tf.createAttributeTerm(created, self);
		addPre(tf.createEqualityTerm(eqc, trueLitTerm));
	    }
	}
	if(!md.getName().equals
	        (ClassInitializeMethodBuilder.CLASS_INITIALIZE_IDENTIFIER)){
	    ProgramVariable ci = 
	        services.getJavaInfo().getAttribute
	        (ImplicitFieldAdder.IMPLICIT_CLASS_INITIALIZED, 
	                services.getJavaInfo().getKeYJavaType(cd));
	    addPre(tf.createEqualityTerm(tf.createVariableTerm(ci), 
					 trueLitTerm));
	}

	services.getImplementation2SpecMap().addMethodSpec(this);
	diverges = null;
	this.params = params;
	this.term2old = term2old;
	this.cSpec = cSpec;
	exc = services.getJavaInfo().getTypeByClassName("Exception");
	excVar = new ProgramVariable(new ProgramElementName(
			       "_exc"+(exceptionVarCount++)), exc);
	nss.programVariables().add(excVar);
	resultVar = makeResultVar();
    }

    protected JMLMethodSpec(MethodDeclaration md){
	this.md = md;
    }

    /**
     * This is used for modelling specification inheritance for overwriten
     * methods.
     */
    public JMLMethodSpec cloneFor(MethodDeclaration md, JMLClassSpec cSpec){
	if(!isCloneableFor(md)) return null;
	JMLMethodSpec cloned = new JMLMethodSpec(md);
	return cloneForHelper(cloned, md, cSpec);
    }

    public JMLMethodSpec copy(){
	JMLMethodSpec copy = 
	    new JMLMethodSpec(md, tf, services, params, 
			      term2old, cSpec, nss);
	return copyHelper(copy);
    }

    public String toString(){
	return inhFrom+"behavior speccase for method "+md.getName()+
	    "\nin context "+cd.getName()+
	    "\nrequires: "+(getPre()==null ? "true" : getPre().toString());
    } 

    public ProgramElement getProofStatement(){
	return createModalityTermForEnsures(trueTerm, false, false, false).
	    javaBlock().program();
    } 

    /**
     * Checks if name, signature, accessibility and so on are equal for 
     * this.md and md. If this.md and md occur on the same branch in the
     * class hierarchy must be checked somewhere else!
     */
    public boolean isCloneableFor(MethodDeclaration md){
	if(!md.getName().equals(this.md.getName())){
	    return false;
	}
	if((md.getTypeReference() != null && 
	    !md.getTypeReference().equals(this.md.getTypeReference())) ||
	   (md.getTypeReference() == null && 
	    this.md.getTypeReference() != null)){
	    return false;
	}
	if((md.getThrown() != null && 
	    !md.getThrown().equals(this.md.getThrown())) ||
	    (md.getThrown() == null && this.md.getThrown() != null)){
	    return false;
	}
	if(md.getParameterDeclarationCount() != 
	       this.md.getParameterDeclarationCount()){
	    return false;
	}
	for(int i = 0; i<md.getParameterDeclarationCount(); i++){
	    if(!md.getParameterDeclarationAt(i).getTypeReference().equals(
		   this.md.getParameterDeclarationAt(i).getTypeReference())){
		return false;
	    }
	}
	if(md.getModifiers().size() != this.md.getModifiers().size() && 
	   !(cd instanceof InterfaceDeclaration)){
	    return false;
	}
	Public p = new Public();
	for(int i = 0; i<md.getModifiers().size(); i++){
	    boolean equal = false;
	    if((cd instanceof InterfaceDeclaration) && 
	       md.getModifiers().getModifier(i).equals(p)){
		continue;
	    }
	    for(int j = 0; j<md.getModifiers().size(); j++){
		if(md.getModifiers().getModifier(i).equals(
		       this.md.getModifiers().getModifier(j))){
		    equal = true;
		    break;
		}
	    }
	    if(!equal) return false;
	}
	return true;
    }

    /**
     * copys several fields to the incomplete copy <code>copy</copy> of this.
     */
    protected JMLMethodSpec copyHelper(JMLMethodSpec copy){
	copy.services = services;
	copy.tf = tf;
	copy.nss = nss;
	copy.params = params;
	copy.pre = pre;
	copy.post = post;
	copy.resultVar = resultVar;
	copy.assignableMode = assignableMode;
	copy.diverges = diverges;
	copy.term2old = (HashMap) term2old.clone();
	copy.assignableVariables = assignableVariables;
	copy.signals = signals;
	copy.excVar = excVar;
	copy.exc = exc;
	return copy;
    }

    protected JMLMethodSpec cloneForHelper(JMLMethodSpec clone, 
					   MethodDeclaration md, 
					   JMLClassSpec cSpec){
	clone.inhFrom = "inherited from "+cd.getName()+" ";
	clone.cSpec = cSpec;
	if(!md.isStatic()){
	    clone.inheritedPrefix = getPrefix();
	    clone.self = tf.createVariableTerm((ProgramVariable) 
					       cSpec.getInstancePrefix());
	}
	clone.cd = cSpec.getClassDeclaration();
	KeYJavaType returnType;
	if(md.getTypeReference() == null){
	    returnType = ji.getKeYJavaType("void");
	}else{
	    returnType = md.getTypeReference().getKeYJavaType();
	}
	clone.pm = 
	    new ProgramMethod(md, ji.getKeYJavaType(cSpec.getClassDeclaration()), 
			      returnType, PositionInfo.UNDEFINED);
	clone = (JMLMethodSpec) copyHelper(clone);
	clone.updateAssignableLocations();
	return clone;
    }

    /**
     * Replaces the inherited prefix with the current prefix in all assignable
     * terms.
     */
    private void updateAssignableLocations(){
	if(!(md.isStatic() || md instanceof Constructor)){
	    SetOfTerm old = assignableVariables;
	    assignableVariables = SetAsListOfTerm.EMPTY_SET;
	    IteratorOfTerm it = old.iterator();
	    while(it.hasNext()){
		assignableVariables = 
		    assignableVariables.add(replaceSelf(it.next()));
	    }
	}
    }

    /**
     * Replaces occurences of the inherited prefix in <code>t</code> with the
     * current prefix.
     */
    private Term replaceSelf(Term t){
	if(t.op() == inheritedPrefix){
	    return self;
	}else if(t.op() instanceof ProgramVariable){
	    return t;
	}else{
	    Term[] subs = new Term[t.arity()];
	    for(int i = 0; i<t.arity(); i++){
		subs[i] = replaceSelf(t.sub(i));
	    }
	    return tf.createTerm(t.op(), subs, 
				 (ArrayOfQuantifiableVariable) null, 
				 null, null);
	}
    }

    public void addSignals(KeYJavaType exc, ProgramVariable v, Term cond){
	signals = signals.add(new Signals(exc, v, cond));
    }

    public SetOfSignals getSignals(){
	return signals;
    }

    public void addPre(Term t){
	if(pre == null){
	    pre = t;
	}else if(t != null){
	    pre = tf.createJunctorTermAndSimplify(Op.AND, pre, t);
	}
    }

    public void addPost(Term t){
	if(post == null){
	    post = t;
	}else if(t != null){
	    post = tf.createJunctorTermAndSimplify(Op.AND, post, t);
	}
    }

    public void addDiverges(Term t){
	if(diverges == null){
	    diverges = t;
	}else if(t != null){
	    diverges = tf.createJunctorTermAndSimplify(Op.OR, diverges, t);
	}
    }

    public void addAssignable(Term t){
	assignableVariables = assignableVariables.add(t);
    }

    public void addAssignable(SetOfTerm as){
	assignableVariables = assignableVariables.union(as);
    }

    public SetOfTerm getAssignable(){
	return assignableVariables;
    }

    public MethodDeclaration getMethodDeclaration(){
	return md;
    }

    public TypeDeclaration getClassDeclaration(){
	return cd;
    }

    /**
     * returns true if this specification demands termination of the method,
     * which means in this case that terminating by throwing an exception
     * is also considered to be a termination.
     */
    public boolean terminates(){
	return diverges == null || diverges == falseTerm;
    }

    /**
     * returns the ProgramVariable that is used for expressing the excetional
     * behavior of a method. 
     */
    public ProgramVariable getExceptionVariable(){
	return excVar;
    }
    
    public SetOfTerm replaceModelFieldsInAssignable(){
	return replaceModelFieldsInAssignable(cSpec);
    }

    public SetOfTerm replaceModelFieldsInAssignable(JMLClassSpec cs){
	if(assignableVariables == SetAsListOfTerm.EMPTY_SET){
	    if("nothing".equals(getAssignableMode())){ 
		return assignableVariables;
	    }/*else if(services.getImplementation2SpecMap().getModifiers(md).
		     contains("pure") && md instanceof ConstructorDeclaration){
		ListOfField l = services.getJavaInfo().getAllFields(
		    (ClassDeclaration) cd);
		IteratorOfField it = l.iterator();
		Term self = tf.createVariableTerm(
		    (ProgramVariable) cSpec.getInstancePrefix());
		while(it.hasNext()){
		    ProgramVariable p = 
			(ProgramVariable) it.next().getProgramVariable();
		    if(!p.isStatic()){
			Term attr = tf.createAttributeTerm(p, self);
			assignableVariables = assignableVariables.add(attr);
		    }
		}
		return assignableVariables;
		}*/else{
		return null;
	    }
	}
	SetOfTerm result = assignableVariables;
	IteratorOfTerm it = assignableVariables.iterator();
	while(it.hasNext()){
	    Term varTerm = it.next();
	    Term rep = (Term) cs.getRepresents().get(varTerm);
	    if(rep != null){
		result = result.remove(varTerm).add(rep);
	    }
	}
	return result;
    }

    public void setAssignableMode(String s){
	assignableMode = s;
    }

    public String getAssignableMode(){
	if(services.getImplementation2SpecMap().getModifiers(md).
	   contains("pure") && 
	   assignableVariables == SetAsListOfTerm.EMPTY_SET){
	    return "nothing";
	}
	if(assignableVariables == SetAsListOfTerm.EMPTY_SET &&
	    assignableMode == null){
	    return "everything";
	}
	// an inconsistent case
	if(assignableVariables != SetAsListOfTerm.EMPTY_SET &&
	   "nothing".equals(assignableMode)){
	    return null;
	}
	return assignableMode;
    }

    /**
     * returns the locally declared precondition
     */
    public Term getPre(){
	return pre;
    }

    /** 
     * returns the precondition + represents for model fields
     * (+ invariants iff withClassSpec = true).
     */
    public Term getCompletePre(boolean withClassSpec, boolean allInv){
	Term t = getPre();
	if(withClassSpec){
	    if(allInv && !services.getImplementation2SpecMap().
	       getModifiers(md).contains("helper")){
		Term ai = UsefulTools.allInvariants(
		    services.getImplementation2SpecMap());
		t = UsefulTools.makeConjunction(t, ai);
	    }else{
		t = cSpec.addClassSpec2Pre(t, md, cSpec); 
	    }
	}
	t = UsefulTools.addRepresents(t, services, cSpec);
	t = updatePrefix(t);
	return t==null? trueTerm : t; 
    }

    /**
     * returns the locally declared postcondition
     */
    public Term getPost(){
	return post;
    }

    /** 
     * returns the postcondition + represents for model fields
     * (+ invariants and constraints iff withClassSpec = true).
     */
    public Term getCompletePost(boolean withClassSpec, boolean allInv){
	Term result;
	if(oldLVs == null){
	    getPi1();
	}
	if(getPost() == null){
	    result = 
		createModalityTermForEnsures(trueTerm, false, withClassSpec, 
					     allInv);
	}else{
	    result = 
		createModalityTermForEnsures(getPost(), false, withClassSpec, 
					     allInv);
	}
	result = result.sub(0);
	result = addOldQuantifiers(result, lv2old, false);
	if(!(md instanceof Constructor)){
	    result = updatePrefix(result);
	}
	return result;
    }

    protected Term updatePrefix(Term t){
	if(!md.isStatic() && t != null && inheritedPrefix != null){
	    t = tf.createUpdateTerm(
		      tf.createVariableTerm((ProgramVariable) inheritedPrefix),
		      self, 
		      t);
	}
	return t;
    }

    public Term getPi1(){
	if(pi1 == null){
	    lv2old = new HashMap();
	    LinkedList l = new LinkedList();
	    pi1 = getPi1Helper(trueTerm, l, term2old);
	    pi1 = getPi1Helper(pi1, l, cSpec.getTerm2Old());
	    pi1 = UsefulTools.addRepresents(pi1, services, cSpec);
	    oldLVs = QuantifierPrefixEntry.toUniversalList ( l.iterator () );
	}
	return pi1;
    }

    private Term getPi1Helper(Term pi1, LinkedList l, HashMap map){
	Iterator it = map.keySet().iterator();
	Term postAndInvTerm = 
	    createModalityTermForEnsures(getPost() == null?
					 trueTerm : getPost(),
					 false, true, false);
	while(it.hasNext()){
	    Term t = (Term) it.next();
	    Term old = (Term) map.get(t);
	    if(UsefulTools.occursIn(old, postAndInvTerm, false)){
		Term t1 = createEqualityTermForOldLV(t, old, l, lv2old);
		if(pi1 == trueTerm){
		    pi1 = t1;
		}else{
		    pi1 = tf.createJunctorTerm(Op.AND,new Term[]{pi1, t1});
		}
	    }
	}
	return pi1;
    }

    public ListOfQuantifierPrefixEntry getOldLVs(){
	if(oldLVs == null){
	    getPi1();
	}
	return oldLVs;
    }

    public HashMap getTerm2Old(){
	return term2old;
    }

    public ProgramVariable getResultVar(){
	if(resultVar == null){
	    resultVar = makeResultVar();
	}
	return resultVar;
    }

    private ProgramVariable makeResultVar(){
	if(md.getTypeReference()==null){
	    return null;
	}
	ProgramVariable v = 
	    new ProgramVariable(new ProgramElementName("_jmlresult"+
						       (resultVarCount++)),
				md.getTypeReference().getKeYJavaType());
	nss.programVariables().add(v);
	return v;
    }

    protected static Term createEqualityTermForOldLV(
	Term t, Term old, LinkedList l, HashMap lv2old){
	TermFactory tf0 = TermFactory.DEFAULT;
	Term lvTerm = null; 
	Term lvEq = null;
	LogicVariable oldlv=null; 
	if(!(old.op() instanceof ProgramVariable) || 
	   old.sort() == Sort.FORMULA && 
	   !(old.sub(0).op() instanceof ProgramVariable)){
	    Term oldFTerm = (t.sort() == Sort.FORMULA ?
			     old.sub(0) : old);
	    t = tf0.createEqualityTerm(old, t);
	    for(int i = 0; i<oldFTerm.arity(); i++){
		LogicVariable arglv = 
		    new LogicVariable(
			new Name(oldFTerm.sub(i).op().name().
				 toString()+
				 "_lv"),
			oldFTerm.sub(i).sort());
		t = UsefulTools.
		createQuantifierTermWithPV(Op.ALL, oldFTerm.sub(i), arglv, t);
	    }
	    return t;
	}else{
	    if(t.sort() != Sort.FORMULA){
		oldlv = 
		    new LogicVariable(new Name("y"+(logicVarCount++)),
				      t.sort());
		l.add(oldlv);
		lvEq = lvTerm = tf0.createVariableTerm(oldlv);
	    }else{
		oldlv = new LogicVariable(
		    new Name("y"+(logicVarCount++)),
		    falseLitTerm.sort());
		l.add(oldlv);
		lvTerm = tf0.createVariableTerm(oldlv);
		lvEq = tf0.createEqualityTerm(lvTerm,trueLitTerm);
		old = old.sub(0);
	    }
	    Term t1 = tf0.createEqualityTerm(lvEq, t);
	    if(lv2old != null) lv2old.put(lvTerm, old);
	    return t1;
	}
    }


    /** Applies the transformation tau_update to <code>a</code> iff 
     * <code>useQuantifiers == false</code> or tau_old otherwise.
     */
    public static Term addOldQuantifiers(Term a, HashMap term2old, 
					 boolean useQuantifiers){
	if(a==null) return a;
	TermFactory tf0 = TermFactory.DEFAULT;
	Iterator it = term2old.keySet().iterator();
	while(it.hasNext()){
	    Term t = (Term) it.next();
	    Term old = (Term) term2old.get(t);
	    if(UsefulTools.occursIn(old, a, false)){
		if(useQuantifiers){
		    LinkedList l = new LinkedList();
		    Term t1 = createEqualityTermForOldLV(t, old, l, null);
		    a = TermFactory.DEFAULT.createJunctorTerm(
			Op.IMP,new Term[]{t1,a});
		    if(t.sort() == Sort.FORMULA){
			old = old.sub(0);
		    }
		    if(old.op() instanceof ProgramVariable){
			a = UsefulTools.createQuantifierTermWithPV(
			    Op.ALL, old, (LogicVariable) l.getFirst(), a);
		    }
		}else{
		    if(t.sort() != Sort.FORMULA && 
		       old.op() instanceof ProgramVariable){
			a = tf0.createUpdateTerm(old,t,a);
		    }else if(t.sort() == Sort.FORMULA && 
			     old.sub(0).op() instanceof ProgramVariable){
			Term oldVTerm = old.sub(0);
			LogicVariable oldlv = 
			    new LogicVariable(
				new Name(oldVTerm.op().name().toString()+
					 "_lv"),
				falseLitTerm.sort());
			Term lvTerm = 
			    tf0.createVariableTerm(oldlv);
			Term t1 = tf0.createEqualityTerm(lvTerm, trueLitTerm);
			t = tf0.createEqualityTerm(t, t1); 
			a = tf0.createJunctorTerm(Op.IMP,new Term[]{t,a});
			a = UsefulTools.createQuantifierTermWithPV(
			    Op.ALL, oldVTerm, oldlv, a);
		    }else{
			t = createEqualityTermForOldLV(t, old, null, null);
			a = tf0.createJunctorTerm(Op.IMP,new Term[]{t,a});
		    }
		}
	    }
	}
	return a;
    }

    public ReferencePrefix getPrefix(){
	if(md != null && md.isStatic()){
	    return cSpec.getStaticPrefix();
	}else{
	    return cSpec.getInstancePrefix();
	}
    }

    public StatementBlock makeMBS(){
	if(mBS == null){
	    ArrayOfParameterDeclaration aopd = md.getParameters();
	    Expression[] aop = new Expression[md.getParameters().size()]; 
	    for(int i=0; i<aopd.size(); i++){
		aop[i] = (ProgramVariable) aopd.
		    getParameterDeclaration(i).
		    getVariableSpecification().getProgramVariable();
	    }
	    if(!(md instanceof Constructor)){
		MethodBodyStatement mb = new MethodBodyStatement(
		    pm,
		    getPrefix(),
		    getResultVar(),
		    new ArrayOfExpression(aop));
		mBS = new StatementBlock(mb);
	    }else{
		New n = new New(aop,
				(TypeReference) cSpec.getStaticPrefix(),
				cSpec.getStaticPrefix());
		mBS = new StatementBlock(new CopyAssignment(
					     (ProgramVariable) getPrefix(),
					     n
					     ));
	    }
	}
	return mBS;
    }

    public ProgramMethod getProgramMethod(){
	return pm;
    }

    protected Term createModalityTermForEnsures(Term post, boolean desugar,
						boolean withInv, 
						boolean allInv){
	Term excVarTerm = tf.createVariableTerm(excVar);
	Term excPost = null;
	post = tf.createJunctorTermAndSimplify(
	    Op.IMP, tf.createEqualityTerm(excVarTerm, nullTerm), post);
	IteratorOfSignals it = signals.iterator();
	while(it.hasNext()){
	    Signals sig = it.next();
	    KeYJavaType excType = sig.type();
	    SortDefiningSymbols s = (SortDefiningSymbols)(excType.getSort());
	    InstanceofSymbol func
		= (InstanceofSymbol)s.lookupSymbol(InstanceofSymbol.NAME);
	    Term excInstTerm = tf.createFunctionTerm(func, excVarTerm);
	    excInstTerm = tf.createEqualityTerm(excInstTerm, trueLitTerm);
	    Term post1 = tf.createJunctorTermAndSimplify(
		Op.IMP,	excInstTerm, sig.condition());
	    if(sig.variable() != null){
		post1 = tf.createUpdateTerm(
		    tf.createVariableTerm(sig.variable()), excVarTerm, post1);
	    }
	    if(excPost != null){
		excPost = tf.createJunctorTermAndSimplify(
		    Op.AND, excPost, post1);
	    }else{
		excPost = post1;
	    }
	}
	JavaBlock jb = catchAllJB(desugar);
	if(excPost != null){
	    excPost = tf.createJunctorTermAndSimplify(
		Op.IMP,	tf.createJunctorTerm(
		    Op.NOT, tf.createEqualityTerm(excVarTerm, nullTerm)),
		excPost);
	    post = tf.createJunctorTermAndSimplify(Op.AND, post, excPost);
	}
	if(withInv && !services.getImplementation2SpecMap().
	   getModifiers(md).contains("helper")){
	    post = cSpec.addClassSpec2Post(post, true, !allInv, md, cSpec); 
	    if(allInv){
		Term ai = UsefulTools.allInvariants(
		    services.getImplementation2SpecMap());
		if(ai != null){
		    post = tf.createJunctorTermAndSimplify(Op.AND , ai, post);
		}
	    }
	}
	post = UsefulTools.addRepresents(post, services, cSpec);
	if(md instanceof Constructor){
	    post = updatePrefix(post);
	}
	//throwing an error is always permitted
/*	KeYJavaType error = services.getJavaInfo().getTypeByClassName("Error");
	SortDefiningSymbols s = (SortDefiningSymbols)(error.getSort());
	InstanceofSymbol func
	    = (InstanceofSymbol)s.lookupSymbol(InstanceofSymbol.NAME);
	Term excInstTerm = tf.createFunctionTerm(func, excVarTerm);
	Term errorTerm = tf.createJunctorTerm(
	    Op.NOT, tf.createEqualityTerm(excVarTerm, nullTerm));
	errorTerm = tf.createJunctorTerm(Op.AND, new Term[]{
	    errorTerm, tf.createEqualityTerm(excInstTerm, trueLitTerm)});
	    post = tf.createJunctorTerm(Op.OR, new Term[]{post, errorTerm});*/

	if(diverges == null || falseTerm.equals(diverges)){
	    return tf.createDiamondTerm(jb, post);
	}else{
	    return tf.createBoxTerm(jb, post);
	}
    }

    protected JavaBlock catchAllJB(boolean desugar){
	JavaBlock catchAllJB;
	if(catchAllStmt == null){
	    ParameterDeclaration param = 
		new ParameterDeclaration(new Modifier[0], new TypeRef(exc),
					 new VariableSpecification(excVar), 
					 false);
	    catchAllStmt = new CatchAllStatement(makeMBS(), param);
	}
	if(desugar){
	    catchAllJB = 
		JavaBlock.createJavaBlock(
		    (StatementBlock) catchAllStmt.desugar());
	}else{
	    catchAllJB = JavaBlock.createJavaBlock(new StatementBlock(
						       catchAllStmt));
	}
	return catchAllJB;
    }

    /**
     * @param t1 the precondition.
     * @param jb the JavaBlock used in the modality.
     */
    protected Term createDiverges(){
	if(diverges != null){
	    Term div = tf.createDiamondTerm(catchAllJB(true), trueTerm);
	    Term divPre = 
		tf.createJunctorTerm(Op.NOT, new Term[]{diverges}); 
	    return tf.createJunctorTermAndSimplify(Op.IMP, divPre, div);
	}
	return trueTerm;
    }
	
    /** Creates a DL formula from this method specification:
     * ((pre & inv) -> [{methodbody}]post) & ((pre & inv & !diverges) -> <{methodbody}>post)
     * (This is the DL formula that is needed for an "ensured postcondition 
     * proofobligation")
     * @param withInvariant: iff true the invariant of the containing type
     * is added to the postcondition which is needed for JMLPostAndInvPOs 
     * created by JMLMethodContracts.
     * @param allInvariants: iff true all existing invariants for every 
     * existing type are added to the precondition (and postcondition if 
     * withInvariant is also true).
     */
    public Term createDLFormula(boolean withPostInvariant, 
				boolean allInvariants){
	Term result,t1,t2;
	result = t2 = null;
	t1 = getPre();
	if(allInvariants && !services.getImplementation2SpecMap().
	   getModifiers(md).contains("helper")){
	    Term ai = UsefulTools.allInvariants(
		services.getImplementation2SpecMap());
	    t1 = UsefulTools.makeConjunction(t1, ai);
	}else{
	    t1 = cSpec.addClassSpec2Pre(t1, md, cSpec);
	}
	t2 = getPost();
	if(t2 == null){
	    result= createModalityTermForEnsures(trueTerm, true, 
						 withPostInvariant,
						 allInvariants);
	}else{
	    result= createModalityTermForEnsures(t2, true,
						 withPostInvariant,
						 allInvariants);
	}
	if(diverges != null){
	    result = tf.createJunctorTermAndSimplify(
		Op.AND,	result, createDiverges()); 
	}
	if(t1 != null){
	    result = tf.createJunctorTermAndSimplify(Op.IMP, t1, result);
	}
	result = addOldQuantifiers(result, term2old, false);
	if(cSpec!=null){
	    result = addOldQuantifiers(result, cSpec.getTerm2Old(), false);
	}
	result = UsefulTools.addRepresents(result, services, cSpec);
	if(UsefulTools.purityCheck(result, 
				   services.getImplementation2SpecMap())
	    != null){
	    throw new RuntimeException("Nonpure method "+ 
				       UsefulTools.purityCheck(result, 
					 services.getImplementation2SpecMap())+
				       " used in the "+
				       "specification for method "+
				       md.getName());
	}
	if(!md.isStatic()){
	    result = updatePrefix(result);
	    result = UsefulTools.quantifySelf(result, md, (ProgramVariable) 
					      getPrefix());
	}
	return UsefulTools.quantifyProgramVariables(result, 
			    params.allElements().iterator(), Op.ALL);
    }

}
