// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.jml;

import java.util.HashMap;

import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.Op;

/** A normal jml methodspecification declared by the jml-keyword 
    <code>normal_behavior</code>*/
public class JMLNormalMethodSpec extends JMLMethodSpec {
    
    public JMLNormalMethodSpec(MethodDeclaration md, 
			 TermFactory tf, Services services, Namespace params, 
			 HashMap term2old, JMLClassSpec cSpec, NamespaceSet nss){
	super(md, tf, services, params, term2old, cSpec, nss);
	addSignals(exc, null, falseTerm);
    }

    protected JMLNormalMethodSpec(MethodDeclaration md){
	super(md);
    }

    /**
     * This is used for modelling specification inheritance for overwriten
     * methods.
     */
    public JMLMethodSpec cloneFor(MethodDeclaration md, JMLClassSpec cSpec){
	if(!isCloneableFor(md)) return null;
	JMLNormalMethodSpec cloned = new JMLNormalMethodSpec(md);
	return (JMLNormalMethodSpec) cloneForHelper(cloned, md, cSpec);
    }

    public ProgramElement getProofStatement(){
	return makeMBS().getStatementAt(0);
    }

    protected Term createModalityTermForEnsures(Term post, boolean desugar,
						boolean withInv, 
						boolean allInv){
	if(diverges == null || falseTerm.equals(diverges)){
	    JavaBlock jb = 
		JavaBlock.createJavaBlock(new StatementBlock(makeMBS()));
	    if(withInv && !services.getImplementation2SpecMap().
	       getModifiers(md).contains("helper")){
		post = cSpec.addClassSpec2Post(post, true, !allInv, md, cSpec);
		if(allInv){
		    Term ai = UsefulTools.allInvariants(
			services.getImplementation2SpecMap());
		    if(ai != null){
			post = 
			    tf.createJunctorTermAndSimplify(Op.AND, ai, post);
		    }
		}
	    }
	    post = UsefulTools.addRepresents(post, services, cSpec);
	    return tf.createDiamondTerm(jb, post);
	}else{
	    Term excVarTerm = tf.createVariableTerm(excVar);
	    post = tf.createJunctorTermAndSimplify(
		Op.AND,	tf.createEqualityTerm(excVarTerm, nullTerm), post);
	    JavaBlock jb = catchAllJB(desugar);
	    if(withInv && !services.getImplementation2SpecMap().
	       getModifiers(md).contains("helper")){
		post = cSpec.addClassSpec2Post(post, true, !allInv, md, cSpec);
		if(allInv){
		    Term ai = UsefulTools.allInvariants(
			services.getImplementation2SpecMap());
		    if(ai != null){
			post = 
			    tf.createJunctorTermAndSimplify(Op.AND, ai, post);
		    }
		}
	    }
	    post = UsefulTools.addRepresents(post, services, cSpec);
	    return tf.createBoxTerm(jb, post);
	}
    }

    public JMLMethodSpec copy(){
	JMLNormalMethodSpec copy = 
	    new JMLNormalMethodSpec(md, tf, services, params, 
				    term2old, cSpec, nss);
	return copyHelper(copy);
    }

    public String toString(){
	return inhFrom+"normal_behavior speccase for method "+md.getName()+
	    " \nin context "+cd.getName()+
	    " \nrequires: "+(getPre()==null ? "true" : getPre().toString());
    } 
}

