// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/*
 * This class is used to create new metavariables, which are also
 * added to the variable namespace.
 */

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.op.Metavariable;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.gui.LDTSettings;
import de.uka.ilkd.key.logic.ldt.NumberLDT;
import java.util.HashMap;

public class MetavariableDeliverer {

    /** Map from "String" to "Integer", containing the highest index
     * number used for a given basename */
    private HashMap  nameCnt = new HashMap ();

    private Proof    proof;


    public MetavariableDeliverer ( Proof p_proof ) {
	proof = p_proof;
    }


    /**
     * Check whether p_sort allows metavariables
     * This is currently hard-coded and only forbidden for the
     * "numbers"-sort
     * @return true iff p_sort allows metavariables
     */
    private static boolean checkSort ( Sort     p_sort ) {
	return NumberLDT.getNumberSymbol().argSort ( 0 ) != p_sort;
    }


    /**
     * Construct a unique identifier by adding a number to the base
     * name
     */
    private Name createNewName ( String p_basename ) {
	Integer cnt = (Integer)nameCnt.get ( p_basename );
	    
	if ( cnt == null )
	    cnt = new Integer ( 0 );
	else
	    cnt = new Integer ( cnt.intValue () + 1 );

	nameCnt.put ( p_basename, cnt );

	return new Name ( p_basename + "_" + cnt );
    }


    /**
     * Create new metavariable and add it to the namespace var_ns of
     * the mediator.
     * @param p_basename Base name of the variable, which will be
     * concatenated with a counter
     * @param p_sort Sort of the variable
     * @return New metavariable, or null iff p_sort allows no
     * metavariables
     */
    public Metavariable createNewVariable ( String p_basename,
					    Sort   p_sort ) {
	if ( !checkSort ( p_sort ) )
	    return null;
 
	Metavariable var = new Metavariable ( createNewName ( p_basename ),
					      p_sort );

	proof.getNamespaces ().variables ().add ( var );

	return var;
    }


    /**
     * Create new variable for instantiation of unknown schema
     * variables when applying taclets.
     * @param p_sort Sort of the variable
     * @return New metavariable, or null iff p_sort allows no
     * metavariables
     */
    public static Metavariable createNewMatchVariable ( String p_basename,
							Sort   p_sort ) {
	if ( !checkSort ( p_sort ) ) return null;

	return Metavariable.createTemporaryVariable ( new Name ( p_basename ),
                                                      p_sort );
    }


}
