// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
/*
 * Created on 14.12.2004
 */
package de.uka.ilkd.key.logic.op;

import java.lang.ref.WeakReference;
import java.math.BigInteger;
import java.util.WeakHashMap;

import de.uka.ilkd.key.logic.Name;

/**
 * @author bubel
 */
public class AnonymousUpdate extends UpdateOperator {

    private static WeakHashMap operatorPool = new WeakHashMap(20);
    private static BigInteger globalCounter = BigInteger.ZERO;
    
    
    /**
     *  
     */
    private AnonymousUpdate(Name name) {
        super(name, new ArrayOfLocation());
    }

    /**
     * returns the anonymous update of the given name or creates a
     * new one if it does not exist
     * @param name
     * @return
     */
    public static AnonymousUpdate getAnonymousOperator(Name name) {        
            WeakReference anonymousUpdateReference = 
                (WeakReference) operatorPool.get(name);
            if (anonymousUpdateReference == null
                    || anonymousUpdateReference.get() == null) {
                // wrapping attributeOp in a weak reference is necessary as
                // it contains a strong reference to its key
                anonymousUpdateReference = 
                       new WeakReference(new AnonymousUpdate(name));
                operatorPool.put(name, anonymousUpdateReference);
            }
            return (AnonymousUpdate) anonymousUpdateReference.get();
        } 
    
    /**
     * returns a new anonymous update 
     * @param name
     * @return
     */
    public static AnonymousUpdate getNewAnonymousOperator() {        
        WeakReference anonymousUpdateReference = null;
        Name name; 
        do { 
             globalCounter = globalCounter.add(BigInteger.ONE);
             name = new Name("*:=*"+globalCounter);
             anonymousUpdateReference = 
                 (WeakReference) operatorPool.get(name);
         } while(anonymousUpdateReference != null);       
        // wrapping attributeOp in a weak reference is necessary as
        // it contains a strong reference to its key
        anonymousUpdateReference = 
            new WeakReference(new AnonymousUpdate(name));
        operatorPool.put(name, anonymousUpdateReference);        
        return (AnonymousUpdate) anonymousUpdateReference.get();
    }    
}
