// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/* Generated by Together */

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ClashFreeSubst;

public class SubstOp extends Op {
    
    SubstOp(Name name) {
	super(name);
    }

    /**
     * @return true iff the sort of the subterm 0 of the given term
     * has the same sort as or a subsort of the variable to be substituted 
     * and the
     * term's arity is 2 and the numer of variables bound there is 0
     * for the 0th subterm and 1 for the 1st subterm.
     */
    public boolean validTopLevel(Term term){
	if (term.arity() != 2) return false;
	if (term.varsBoundHere(1).size() != 1) return false;
	if (term.varsBoundHere(0).size() != 0) return false;
	Sort substSort=term.sub(0).sort();
	Sort varSort=term.varsBoundHere(1).getQuantifiableVariable(0).sort();
	return substSort.extendsTrans(varSort);
    }

    /**
     * @return sort of the second subterm or throws an
     * IllegalArgumentException if the given term has no correct (2=) arity
     */
    public Sort sort(Term[] term) {
	if (term.length==2)
	    return term[1].sort();
	else throw new IllegalArgumentException("Cannot determine sort of "+
						"invalid term (Wrong arity).");
    }

    /** @return arity of the Substitution operator as int */
    public int arity() {
	return 2;
    }

    /**
     * Apply this substitution operator to <code>term</code>, which
     * has this operator as top-level operator
     */
    public Term apply ( Term term ) {
	QuantifiableVariable v=term.varsBoundHere(1).getQuantifiableVariable(0);
	ClashFreeSubst cfSubst = new ClashFreeSubst(v, term.sub(0));
	Term res = cfSubst.apply(term.sub(1));
	return res;
    }
}
