// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op.oclop;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.oclsort.*;

public abstract class OclOp {

    public static final Function INVARIANT = new Function(new Name("$invariant"),
							  OclSort.OCLINVARIANT,
							  new Sort[]{OclSort.OCLTYPE,
								     OclSort.BOOLEAN});

    public static final Function CONS_INV = new Function(new Name("$cons_inv"),
							 OclSort.SET_OF_OCLINVARIANT,
							 new Sort[]{OclSort.OCLINVARIANT,
								    OclSort.SET_OF_OCLINVARIANT});

    public static final Function NIL_INV = new Function(new Name("$nil_inv"),
							OclSort.SET_OF_OCLINVARIANT,
							new Sort[0]);
    
    public static final OclInsertSet INSERT_SET = new OclInsertSet();
    public static final OclInsertBag INSERT_BAG = new OclInsertBag();
    public static final OclInsertSequence INSERT_SEQUENCE = new OclInsertSequence();
    public static final OclEmptySet EMPTY_SET = new OclEmptySet();
    public static final OclEmptyBag EMPTY_BAG = new OclEmptyBag();
    public static final OclEmptySequence EMPTY_SEQUENCE = new OclEmptySequence();
    public static final OclIterate ITERATE = new OclIterate();
    public static final OclQuantifier FOR_ALL = new OclQuantifier(new Name("$forAll"));
    public static final OclQuantifier EXISTS = new OclQuantifier(new Name("$exists"));
    public static final OclQuantifier ONE = new OclQuantifier(new Name("$one"));
    public static final OclIsUnique IS_UNIQUE = new OclIsUnique();
    public static final OclSortedBy SORTED_BY = new OclSortedBy();
    public static final OclAny ANY = new OclAny();
    public static final OclSetOp INCLUDING = new OclSetOp(new Name("$including"));
    public static final OclSetOp EXCLUDING = new OclSetOp(new Name("$excluding"));
    public static final OclSequenceOp APPEND = new OclSequenceOp(new Name("$append"));
    public static final OclSequenceOp PREPEND = new OclSequenceOp(new Name("$prepend"));
    public static final OclCollectionConversion AS_SET 
	= new OclCollectionConversion(new Name("$asSet"), CollectionSort.SET);
    public static final OclCollectionConversion AS_BAG 
	= new OclCollectionConversion(new Name("$asBag"), CollectionSort.BAG);
    public static final OclCollectionConversion AS_SEQUENCE 
	= new OclCollectionConversion(new Name("$asSequence"), CollectionSort.SEQUENCE);
    public static final OclSubSequence SUB_SEQUENCE = new OclSubSequence();
    public static final OclSequenceElems AT = new OclSequenceElems(new Name("$at"), 2);
    public static final OclSequenceElems FIRST = new OclSequenceElems(new Name("$first"), 1);
    public static final OclSequenceElems LAST = new OclSequenceElems(new Name("$last"), 1);
    public static final OclIf IF = new OclIf();
    public static final Function ALL_SUBTYPES = new Function(new Name("$allSubtypes"),
							     //new CollectionSort(CollectionSort.SET, 
							     //		       OclSort.OCLTYPE),
							     OclSort.SET_OF_OCLTYPE,
							     new Sort[]{OclSort.OCLTYPE});
    public static final Function ALL_INSTANCES = new Function(new Name("$allInstances"),
							      //new CollectionSort(CollectionSort.SET, 
							      //		   OclSort.CLASSIFIER),
							      OclSort.SET_OF_CLASSIFIER,
							      new Sort[]{OclSort.OCLTYPE});
    public static final OclUnion UNION = new OclUnion();
    public static final OclIntersection INTERSECTION = new OclIntersection();
    public static final OclDifference DIFFERENCE = new OclDifference(new Name("$difference"));
    public static final OclDifference SYMMETRIC_DIFFERENCE 
	= new OclDifference(new Name("$symmetricDifference"));
    public static final OclSubsetOp SELECT = new OclSubsetOp(new Name("$select"));
    public static final OclSubsetOp REJECT = new OclSubsetOp(new Name("$reject"));
    public static final OclCollect COLLECT = new OclCollect();
    public static final Function AND = new Function(new Name("$and"),
						    OclSort.BOOLEAN,
						    new Sort[]{OclSort.BOOLEAN, OclSort.BOOLEAN});
    public static final Function OR = new Function(new Name("$or"),
						   OclSort.BOOLEAN,
						   new Sort[]{OclSort.BOOLEAN, OclSort.BOOLEAN});
    public static final Function XOR = new Function(new Name("$xor"),
						    OclSort.BOOLEAN,
						    new Sort[]{OclSort.BOOLEAN, OclSort.BOOLEAN});
    public static final Function IMPLIES 
	= new Function(new Name("$implies"),
		       OclSort.BOOLEAN,
		       new Sort[]{OclSort.BOOLEAN, OclSort.BOOLEAN});
    public static final Function NOT = new Function(new Name("$not"),
						    OclSort.BOOLEAN,
						    new Sort[]{OclSort.BOOLEAN});
    public static final Function EQUALS = new Function(new Name("$equals"),
						       OclSort.BOOLEAN,
						       new Sort[]{OclSort.OCLGENERIC, 
								  OclSort.OCLGENERIC});
    public static final Function NOT_EQUALS = new Function(new Name("$notEquals"),
							   OclSort.BOOLEAN,
							   new Sort[]{OclSort.OCLGENERIC, 
								      OclSort.OCLGENERIC});
    public static final Function OCL_IS_NEW = new Function(new Name("$oclIsNew"),
							   OclSort.BOOLEAN,
							   new Sort[]{OclSort.CLASSIFIER});
    public static final Function TRUE = new Function(new Name("$true"),
						     OclSort.BOOLEAN,
						     new Sort[0]);
    public static final Function FALSE = new Function(new Name("$false"),
						      OclSort.BOOLEAN,
						      new Sort[0]);
    public static final Function SELF = new Function(new Name("$self"),
						     OclSort.CLASSIFIER,
						     new Sort[0]);
    public static final Function INCLUDES = new Function(new Name("$includes"),
							 OclSort.BOOLEAN,
							 new Sort[]{OclSort.COLLECTION_OF_OCLANY,
								    OclSort.OCLANY});
    public static final Function EXCLUDES = new Function(new Name("$excludes"),
							 OclSort.BOOLEAN,
							 new Sort[]{OclSort.COLLECTION_OF_OCLANY,
								    OclSort.OCLANY});
    public static final Function COUNT = new Function(new Name("$count"),
						      OclSort.INTEGER,
						      new Sort[]{OclSort.COLLECTION_OF_OCLANY,
								 OclSort.OCLANY});
    public static final Function INCLUDES_ALL 
	= new Function(new Name("$includesAll"),
		       OclSort.BOOLEAN,
		       new Sort[]{OclSort.COLLECTION_OF_OCLANY,
				  OclSort.COLLECTION_OF_OCLANY});
    public static final Function EXCLUDES_ALL 
	= new Function(new Name("$excludesAll"),
		       OclSort.BOOLEAN,
		       new Sort[]{OclSort.COLLECTION_OF_OCLANY,
				  OclSort.COLLECTION_OF_OCLANY});
    public static final Function SIZE = new Function(new Name("$size"),
						     OclSort.INTEGER,
						     new Sort[]{OclSort.COLLECTION_OF_OCLANY});
    public static final Function IS_EMPTY = new Function(new Name("$isEmpty"),
							 OclSort.BOOLEAN,
							 new Sort[]{OclSort.COLLECTION_OF_OCLANY});
    public static final Function NOT_EMPTY = new Function(new Name("$notEmpty"),
							  OclSort.BOOLEAN,
							  new Sort[]{OclSort.COLLECTION_OF_OCLANY});
    public static final Function SUM = new Function(new Name("$sum"),
						    OclSort.REAL,
						    new Sort[]{OclSort.COLLECTION_OF_REAL});

    public static final Function LESS_THAN = new Function(new Name("$lessThan"),
							  OclSort.BOOLEAN,
							  new Sort[]{OclSort.REAL,
								     OclSort.REAL});
    public static final Function LESS_THAN_EQ = new Function(new Name("$lessThanEq"),
							     OclSort.BOOLEAN,
							     new Sort[]{OclSort.REAL,
									OclSort.REAL});
    public static final Function GREATER_THAN = new Function(new Name("$greaterThan"),
							     OclSort.BOOLEAN,
							     new Sort[]{OclSort.REAL,
									OclSort.REAL});
    public static final Function GREATER_THAN_EQ = new Function(new Name("$greaterThanEq"),
								OclSort.BOOLEAN,
								new Sort[]{OclSort.REAL,
									   OclSort.REAL});
    public static final Function PLUS = new Function(new Name("$plus"),
						     OclSort.REAL,
						     new Sort[]{OclSort.REAL,
								OclSort.REAL});
    public static final Function MINUS = new Function(new Name("$minus"),
						      OclSort.REAL,
						      new Sort[]{OclSort.REAL,
								 OclSort.REAL});
    public static final Function TIMES = new Function(new Name("$times"),
						      OclSort.REAL,
						      new Sort[]{OclSort.REAL,
								 OclSort.REAL});
    public static final Function DIV_INFIX = new Function(new Name("$divInfix"),
							  OclSort.REAL,
							  new Sort[]{OclSort.REAL,
								     OclSort.REAL});
    public static final Function DIV = new Function(new Name("$div"),
						    OclSort.INTEGER,
						    new Sort[]{OclSort.INTEGER,
							       OclSort.INTEGER});
    public static final Function MOD = new Function(new Name("$mod"),
						    OclSort.INTEGER,
						    new Sort[]{OclSort.INTEGER,
							       OclSort.INTEGER});
    public static final Function MIN = new Function(new Name("$min"),
						    OclSort.REAL,
						    new Sort[]{OclSort.REAL,
							       OclSort.REAL});
    public static final Function MAX = new Function(new Name("$max"),
						    OclSort.REAL,
						    new Sort[]{OclSort.REAL,
							       OclSort.REAL});
    public static final Function MINUS_PREFIX = new Function(new Name("$plus"),
							     OclSort.REAL,
							     new Sort[]{OclSort.REAL});
    public static final Function ABS = new Function(new Name("$abs"),
						    OclSort.REAL,
						    new Sort[]{OclSort.REAL});

    public static final Function OCL_WRAPPER = new Function(new Name("$OclWrapper"),
							    Sort.FORMULA,
							    new Sort[]{OclSort.OCLGENERIC});
}
