// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.sort;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.op.MapAsListFromNameToSortDependingSymbol;
import de.uka.ilkd.key.logic.op.MapFromNameToSortDependingSymbol;
import de.uka.ilkd.key.logic.op.ListOfSortDependingSymbol;
import de.uka.ilkd.key.logic.op.SLListOfSortDependingSymbol;
import de.uka.ilkd.key.logic.op.IteratorOfSortDependingSymbol;
import de.uka.ilkd.key.logic.op.SortDependingSymbol;


public abstract class AbstractNonCollectionSort extends de.uka.ilkd.key.logic.sort.AbstractSort 
    implements de.uka.ilkd.key.logic.sort.NonCollectionSort {


    protected de.uka.ilkd.key.logic.sort.CollectionSort setSort;

    protected de.uka.ilkd.key.logic.sort.CollectionSort bagSort;

    protected de.uka.ilkd.key.logic.sort.CollectionSort sequenceSort;
    

    private MapFromNameToSortDependingSymbol definedSymbols =
	MapAsListFromNameToSortDependingSymbol.EMPTY_MAP;

    private boolean symbolsCreated = false;


    /** creates a Sort (with a new equality symbol for this sort) */
    public AbstractNonCollectionSort(Name name) {
	super(name);
	setSort=AbstractCollectionSort.createSetSort(this);
	bagSort=AbstractCollectionSort.createBagSort(this);
	sequenceSort=AbstractCollectionSort.createSequenceSort(this);
    }


    public de.uka.ilkd.key.logic.sort.CollectionSort getSetSort() {
	return setSort;
    }

    public de.uka.ilkd.key.logic.sort.CollectionSort getBagSort() {
	return bagSort;
    }

    public de.uka.ilkd.key.logic.sort.CollectionSort getSequenceSort() {
	return sequenceSort;
    }


    /**
     * Create the symbols defined by this sort, insert them into the
     * namespace "p_func_ns"
     * @param p_func_ns Namespace to which functions and predicates
     * should be added to
     */
    public void createSymbols ( Namespace     p_func_ns,
				Namespace     sort_ns) {
	if ( !symbolsCreated ) {

	    ListOfSortDependingSymbol l0 =
		SLListOfSortDependingSymbol.EMPTY_LIST;
	    ListOfSortDependingSymbol l1;

	    l1 = getSetSort      ().createSymbols ();
	    if ( l1 == null )
		return;
	    l0 = l0.prepend ( l1 );

	    l1 = getBagSort      ().createSymbols ();
	    if ( l1 == null )
		return;
	    l0 = l0.prepend ( l1 );

	    l1 = getSequenceSort ().createSymbols ();
	    if ( l1 == null )
		return;
	    l0 = l0.prepend ( l1 );

	    addSymbols ( l0, p_func_ns );

	    symbolsCreated = true;
	}
    }


    /**
     * Lookup the symbol of kind "p_name", which is a sort
     * independent identifier for this symbol
     * @return Symbol with (kind) name "p_name"
     * ("ret.getKind().equals(p_name)"), null if no such object exists
     */
    public SortDependingSymbol lookupSymbol  ( Name      p_name ) {
	return definedSymbols.get ( p_name );
    }


    protected void addSymbols ( ListOfSortDependingSymbol p,
				Namespace                 p_func_ns ) {
	IteratorOfSortDependingSymbol it = p.iterator ();
	SortDependingSymbol           s;
	while ( it.hasNext () ) {
	    s              = it.next ();
	    definedSymbols = definedSymbols.put ( s.getKind (), s );
	    p_func_ns.add ( (Named)s );
	}
    }

}
