// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.sort;

import java.io.ObjectStreamException;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.op.Equality;
import de.uka.ilkd.key.logic.op.Op;

public abstract class AbstractSort implements Sort {

   
    /** name of the Sort */
    protected Name name;

    /** equality symbol for this sort
     * @label equality symbol for sort*/
    protected final Equality equality=Op.EQUALS;

    
    /** creates a Sort (with a new equality symbol for this sort) */
    public AbstractSort(Name name) {	
	this.name=name;
    }

    /**
     * @return the sorts of the successors of this sort
     */
    public abstract SetOfSort extendsSorts();

    /**
     * returns true iff given sort is a part of the transitive closure of
     * the supersorts of this sort. One might optimize by hashing %%
     */
    public boolean extendsTrans(de.uka.ilkd.key.logic.sort.Sort sort) {
	if (sort == this ) return true;
	if (! (sort instanceof ObjectSort ||
	       sort instanceof GenericSort) ) return false;
	IteratorOfSort it = extendsSorts().iterator();	
	while (it.hasNext()) {
	    Sort s = it.next();
	    if (s == null) continue;
	    if (s == sort || s.extendsTrans(sort)) {
		return true;
	    }
	}
	return false;
    }

    protected Object readResolve() 
	throws ObjectStreamException {	

	if (name().equals(Sort.FORMULA.name()) && equality==null) {
	    return Sort.FORMULA;
	} 
	return this;
    }

    
    /** @return name of the Sort as String */
    public Name name() {
	return name;
    }
    
    /** @return equality symbol of this sort */
    public Equality getEqualitySymbol(){
	return (this==Sort.FORMULA ? Op.EQV : equality);
    }
    
    public String toString() {
       return name().toString();
    }



}
