// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.parser;

import antlr.*;

public class NotDeclException extends antlr.SemanticException {
    String cat;
    String filename;
    String undeclared_symbol;
    
    public NotDeclException(String cat, Token t, String filename) {
	super("NotDeclared");
	this.cat      = cat;
	this.filename = filename;
	this.undeclared_symbol = t.getText();
	this.line     = t.getLine();
	this.column   = t.getColumn();
    }

    public NotDeclException(String cat, String undeclared_symbol, 
			    String filename, int line, int column) {
	super("NotDeclared");
	this.filename = filename;
	this.cat      = cat;
	this.undeclared_symbol = undeclared_symbol;
	this.line     = line;
	this.column   = column;
    }

    /**
     * Returns a clean error message (no line number/column information)
     */
    public String getErrorMessage ()
    {
	String errmsg = cat + "\n\t" + undeclared_symbol + "\n";
	return errmsg + "not declared";
    }

    /**
     * Returns a clean error message (no line number/column information)
     */
    public String getMessage () {
	return getErrorMessage();
    }
    
    /**
     * Returns a string representation of this exception.
     */
    public String toString() {
	return filename+"("+this.getLine()+", "+this.getColumn()+"): "
	    +getErrorMessage();
    }
}
