// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.util.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.sort.*;  
    import de.uka.ilkd.key.logic.sort.oclsort.OclSort;
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.rule.metaconstruct.*;
    import de.uka.ilkd.key.rule.*;
    import de.uka.ilkd.key.proof.init.Includes;
    import java.util.List;
    import java.util.LinkedList;
    import java.util.ArrayList;
    import java.util.HashMap;
    import java.util.HashSet;
    import java.util.Iterator;
    import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
    import de.uka.ilkd.key.java.Recoder2KeY;
    import de.uka.ilkd.key.java.JavaReader;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.java.Services;
    import de.uka.ilkd.key.java.recoderext.*;
    import de.uka.ilkd.key.java.abstraction.KeYJavaType;
    import de.uka.ilkd.key.java.abstraction.ListOfKeYJavaType;
    import de.uka.ilkd.key.java.abstraction.SLListOfKeYJavaType;
    import de.uka.ilkd.key.proof.UseInsteadMap;
    import de.uka.ilkd.key.proof.RuleSource;
    import de.uka.ilkd.key.pp.AbbrevMap;
    
}

/** ONLY FOR TEST CASES.
 * A parser for first-order terms and formulae.  Variables in 
 * quantifiers are expected to be declared globally in the variable
 * namespace.  This parser is used for test cases, where you want to know
 * in advance which objects will represent bound variables.
 */  

class GlobalDeclarationTermParser extends TermParser;
options {
	importVocab=KeYLexer;
	k = 1;
	defaultErrorHandler=true;
}


{
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];
    
    protected NamespaceSet nss;

    protected Choice defaultChoice = null; 
    protected HashMap category2Default = new HashMap();
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();

    protected AbbrevMap scm = new AbbrevMap();

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_rulesets;
    protected UseInsteadMap useInstead=new UseInsteadMap();

    protected KeYExceptionHandler keh = null;

    private boolean schemaMode;

    protected JavaReader javaReader;
    protected TermFactory tf;
    private Services services = new Services();

    /** collects errors and warnings that occured during parsing */
    protected ErrorHandler errHandler = new ErrorHandler();

    public boolean errors = false;



    public GlobalDeclarationTermParser(TokenStream lexer,
				       TermFactory tf,
		  		       JavaReader jr,
				       NamespaceSet nss,
                       Choice defaultChoice) {
        this(lexer,2);
        this.tf = tf;
        this.javaReader = jr;
        this.nss     = nss;
        switchToNormalMode();
        this.defaultChoice = defaultChoice;
        this.keh = services.getExceptionHandler();
    }

    public GlobalDeclarationTermParser(TokenStream lexer,
				       TermFactory tf,
		  		       Services services,
				       NamespaceSet nss,
                       Choice defaultChoice) {
	this(lexer, tf, 
		new Recoder2KeY(new KeYCrossReferenceServiceConfiguration(services.getExceptionHandler()), 
				services.getJavaInfo().rec2key(), new NamespaceSet(), 
				services.getTypeConverter()),
   	        nss, defaultChoice);
        this.keh = services.getExceptionHandler();
    }

    public GlobalDeclarationTermParser(TokenStream lexer,
				       Services services,
				       NamespaceSet nss) {
	this(lexer, TermFactory.DEFAULT,
	     new Recoder2KeY(new KeYCrossReferenceServiceConfiguration(services.getExceptionHandler()), 
		 services.getJavaInfo().rec2key(), new NamespaceSet(), services.getTypeConverter()),
	        nss,new Choice(new Name("Default")));
    }
    
    protected boolean isPackage(String name){
        try {
            return getJavaInfo().isPackage(name);
        } catch(RuntimeException e){	    
	    // may be thrown in cases of invalid java identifiers
            return false;
        } 
    }

    private void switchToNormalMode() {
        schemaMode   = false;
    }

    private boolean inSchemaMode() {
	return schemaMode;
    }

    private void switchToSchemaMode() {
        schemaMode   = true;
    }

    public NamespaceSet namespaces(){
    return nss;
    }

    public Namespace sorts() {
	return nss.sorts();
    }

    public Namespace variables() {
	return nss.variables();
    }

    public Namespace functions() {
	return nss.functions();
    }

    public Namespace programVariables() {
        return nss.programVariables();
    }

    public Namespace ruleSets() {
	return nss.ruleSets();
    }

    public Namespace choices() {
	return nss.choices();
    }

    public Named lookup(Name n) {
	final Namespace[] lookups = {
		functions(), variables(), 
		programVariables()
	};

	for (int i = 0; i<lookups.length; i++) {
	    if (lookups[i].lookup(n) != null) {
		return lookups[i].lookup(n);
            }
	}
	return null;	
    }


    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }


    protected int getLine() {
	int line = -1;
	try {
	    line = LT(0).getLine();
	} 
	catch (TokenStreamException e) {
	    System.err.println("No further token in stream");
	}
	return line;
    }   

    protected int getColumn() {
	int col = -1;
	try {
	    col = LT(0).getColumn();
	} 
	catch (TokenStreamException e) {
	    System.err.println("No further token in stream");
	}
	return col;
    }   

    // These three methods are required because the unused productions
    // from terms.g are inherited.  They should not be called.
    protected LogicVariable bindVar(Token id,Sort s) {
	throw new RuntimeException("bindVar was called in " +
				   "GlobalDeclarationTermParser.");
    }
	
    protected ProgramVariable progVar(Token id,Sort s) {
	ProgramVariable v = (ProgramVariable)(ProgramVariable)
		   programVariables().lookup(new Name(id.getText()));
	if (v == null) {
	    v = new ProgramVariable(new ProgramElementName(id.getText()), s);
	    nss.setProgramVariables(programVariables().extended(v));
	}
	return v;
    }

    protected void unbindVars() {
	throw new RuntimeException("unbindVars was called in " +
				   "GlobalDeclarationTermParser.");
    }


    protected static HashSet progVars(JavaBlock jb) {
	ProgramVariableCollector pvc
	    = new ProgramVariableCollector(jb.program());
	pvc.start();
	return pvc.result();
    }


    protected void bindProgVars(HashSet progVars) {
	nss.setProgramVariables(new Namespace(programVariables()));
	Iterator it=progVars.iterator();
	while (it.hasNext()) {
	    programVariables().add((ProgramVariable)it.next());
	}
    }

    protected void unbindProgVars() {
	nss.setVariables(variables().parent());
    }


    protected Term termForParsedVariable(ParsableVariable v) 
	throws antlr.SemanticException 
	{
	    if ( v instanceof LogicVariable ) {
		return tf.createVariableTerm((LogicVariable)v);
	    } else if ( v instanceof ProgramVariable ) {
		return tf.createVariableTerm((ProgramVariable)v);
	    } else {
		throw new KeYSemanticException(
						  ""+ v +" is not a logic variable");
	    }
	}

    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                   filename+"\n " +e);
            }
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }  

    public Term createAttributeTerm(Term prefix, 
				    TermSymbol attribute,
				    int shadowPrimes) {
	Term result = prefix;
	if (((ProgramVariable)attribute).isStatic()){
	    result = tf.createVariableTerm((ProgramVariable)attribute);
	} else {
	    if (shadowPrimes > 0) {
		Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
		result = tf.createShadowAttributeTerm((ProgramVariable)attribute, 
						      result, shadownum);
	    } else {
		result = tf.createAttributeTerm((ProgramVariable)attribute, result);
	    }
	}	
	return result;
    }

    public TermSymbol getAttribute(Sort prefixSort, 
				   String attributeName) 
	throws antlr.SemanticException {
	    final JavaInfo javaInfo = getJavaInfo();
	    TermSymbol v = null;
            if (!inSchemaMode()) {
		if (attributeName.indexOf(':') != -1) {
		   v = javaInfo.getAttribute(attributeName, 
			javaInfo.getKeYJavaType(attributeName.substring(0, attributeName.indexOf(':'))));
	        } else {
		    v = javaInfo.getAttribute(attributeName, (ObjectSort)prefixSort);
                }
            } else {
                v = (SortedSchemaVariable)variables().lookup(new Name(attributeName));
            }
            if ( v == null && !"length".equals(attributeName) ) {
                throw new NotDeclException
                    ("Attribute ", attributeName, getFilename(), 
                     getLine(), getColumn());

            }

	return v;
    }

    public de.uka.ilkd.key.logic.op.Location[] extractLocations(List locNames)
        throws antlr.SemanticException {
	de.uka.ilkd.key.logic.op.Location[] vars = 
	    new de.uka.ilkd.key.logic.op.Location[locNames.size()];
	for (int i = 0; i<vars.length; i++) {
	    final String locName = (String)locNames.get(i);
	    Object o = programVariables().lookup(new Name(locName));
	    if (o != null) {
            vars[i] = (de.uka.ilkd.key.logic.op.Location) o;
	    } else {
		    vars[i] = (de.uka.ilkd.key.logic.op.Location) 
                getAttribute(null, locName);
	    }
	}
	return vars;
    }

    protected JavaBlock getJavaBlock(Token t) {
	return javaReader.readBlockWithEmptyContext(t.getText());
    }

    /** returns the java information object */
    public JavaInfo getJavaInfo() {
	return null;
//	return javaReader.createJavaInfo(new Services());
    }

	public Services getServices() {
	    return services;
    	}
	
       
       public void schema_var_decl(String name,Sort s,
                                    boolean makeVariableSV,
                                    boolean makeDependingSV,
                                    boolean rigid,
                                    boolean listModifier)
	{
              if (!skip_schemavariables) {
                  SchemaVariable v;
                  if ( s == Sort.FORMULA ) {
                      v = SchemaVariableFactory.createFormulaSV
                          (new Name(name), listModifier, rigid);
                      } else if ( s instanceof ProgramSVSort ) {
                          v = SchemaVariableFactory.createProgramSV
                         (new ProgramElementName(name),(ProgramSVSort) s, listModifier);
                      } else {
                        if ( makeVariableSV ) {
                             v=SchemaVariableFactory.createVariableSV
                             (new Name(name), s, listModifier);
                        } else if ( makeDependingSV ) {
                               v = SchemaVariableFactory.createDependingSV
                             (new Name(name), s, listModifier);
                        } else { v = SchemaVariableFactory.createTermSV
                              (new Name(name), s, listModifier, rigid);
                        }
                     }
                     if (inSchemaMode()) {
                         variables().add(v);
                     }
               }

         }


    protected Operator lookupVarfuncId(String varfunc_name, 
                    PairOfTermArrayAndBoundVarsArray args) throws NotDeclException {
        Term[] argTerms = null;
        if (args != null) {
            argTerms = args.getTerms();
        }
        return lookupVarfuncId(varfunc_name, argTerms);
    }

    /** looks up a function, (program) variable or static query of the 
     * given name varfunc_id and the argument terms args in the namespaces 
     * and java info. 
     * @param args is null iff no argument list is given, for instance `f', 
     * and is an array of size zero, if an empty argument list was given,
     * for instance `f()'.
     */
    protected Operator lookupVarfuncId(String varfunc_name, Term[] args) 
                                                    throws NotDeclException{
        // case 1: variable
        Operator v = (TermSymbol) variables().lookup(new Name(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        // case 2: function
        v = (TermSymbol) functions().lookup(new Name(varfunc_name));
        if (v != null) { // we allow both args==null (e.g. `c')
                         // and args.length=0 (e.g. 'c())' here 
            return v;
        }
        // case 3: static query
        if (varfunc_name.lastIndexOf("::")>0 && args!=null) {
            String classname = varfunc_name.substring
            (0, varfunc_name.lastIndexOf("::"));
            varfunc_name = varfunc_name.substring
            (varfunc_name.lastIndexOf("::")+2);
            v = getServices().getJavaInfo().getProgramMethodTerm
                (null, varfunc_name, args, classname).op();
            if (v != null) {
                return v;
            }
        }
        // case 4: program variable
        v = (ProgramVariable) programVariables().lookup
            (new ProgramElementName(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        if (args==null) {
            throw new NotDeclException
            ("(program) variable or constant", varfunc_name,
                getFilename(), getLine(), getColumn());
        } else {
            throw new NotDeclException
            ("function or static query", varfunc_name,
                getFilename(), getLine(), getColumn());
        }
    }

}


top {Term t;} : t=formula;

bindingterm returns [Term a = null]
	:
		a = global_bindingterm
	;

/** here is the special behaviour, that requires a bound variable
 * to be previously declared in <code>variables()</code>. */

global_bindingterm returns [Term a = null]
{
	Operator op = null;
	ParsableVariable v0, v1;
	LogicVariable v = null;
	ProgramVariable pv = null;
	Term a1,a2;
	boolean waryAll = false;
	boolean waryEx = false;
}
	:
		( 
			"all" { op = Op.ALL; }
		|   "ex"  { op = Op.EX;  }
		)
		
		v0=varId DOT // Don't need the sort here
		{v = (LogicVariable) v0;} a1=term60 
		{ 
			a = tf.createQuantifierTerm((Quantifier)op, v,a1);
		}
	|
		LBRACE v1=varId
		(
		a1=term // Don't need the sort here
		( "waryAll" { waryAll = true; } |
		  "waryEx"  { waryEx  = true; } )?
		RBRACE
		{v = (LogicVariable) v1;} a2=term60
		{ 
		    if ( waryAll )
		        a = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
		                                        v, a1, a2 );
		    else if ( waryEx )
		        a = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
		                                        v, a1, a2 );
		    else
		        a = tf.createSubstitutionTerm ( Op.SUBST,
		                                        v, a1, a2 );
		}
		|
		// simple updates, just for tests
		ASSIGN a1=term RBRACE
		{pv = (ProgramVariable) v1;} a2=term60
		{ 
		    a = tf.createUpdateTerm ( tf.createVariableTerm ( pv ), a1, a2 );
		}
		)
	;
