// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser.jml;

	import de.uka.ilkd.key.jml.*;
	
    import java.io.*;
    import antlr.*;
    
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.sort.*;
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.rule.*;
    import de.uka.ilkd.key.collection.*;
    import de.uka.ilkd.key.proof.RuleSource;
    import de.uka.ilkd.key.proof.init.Includes;
    import de.uka.ilkd.key.util.ExtList;

    import de.uka.ilkd.key.rule.metaconstruct.*;
    import de.uka.ilkd.key.java.abstraction.*;
    import de.uka.ilkd.key.java.*;
    import de.uka.ilkd.key.java.declaration.*;
    import de.uka.ilkd.key.java.declaration.modifier.*;
    import de.uka.ilkd.key.java.statement.LoopStatement;
    import de.uka.ilkd.key.java.statement.BranchStatement;
    import de.uka.ilkd.key.parser.NotDeclException; 
    import de.uka.ilkd.key.parser.KeYSemanticException;
    import de.uka.ilkd.key.java.abstraction.KeYJavaType;
    import de.uka.ilkd.key.java.abstraction.Type;
    import de.uka.ilkd.key.java.expression.literal.BooleanLiteral;
    import de.uka.ilkd.key.java.reference.ReferencePrefix;
    import de.uka.ilkd.key.java.reference.TypeReference;
    import de.uka.ilkd.key.java.reference.TypeRef;

    import java.util.Vector;
    import java.util.LinkedList;
    import java.util.HashSet;
    import java.util.Set;
    import java.util.Iterator;
    import java.util.HashMap;
    import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
    import de.uka.ilkd.key.java.Recoder2KeY;
    import de.uka.ilkd.key.java.Services;
    import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
}


class KeYJMLParser extends Parser;
options {
	importVocab=KeYJMLLexer;
	k = 1;
	defaultErrorHandler=true;
}

{
    private static int oldVarCount = 0;
    private static int paramVarCount = 0;
    private static int mmCount = 0;
    private ArithOpProvider aOP;
    private SetOfJMLMethodSpec specs = SetAsListOfJMLMethodSpec.EMPTY_SET;
    private TermFactory tf;
    private NamespaceSet nss;
    private TypeDeclaration cld;
    private MethodDeclaration md;
    private JMLClassSpec cSpec;
    //the prefix needed for attribute terms.
    private ReferencePrefix prefix = null;
    //the namespace for the methodparameters.
    private Namespace param_ns = new Namespace();
    //mapping of expression to a term representing their evaluation in the pre
    //state. Needed for the \old construct.
    private HashMap term2old = new HashMap();
    //used for caching of the result variable delivered by JMLMethodSpec
    private ProgramVariable result;
    //used to cache the KeYJavaType of a term. Ugly, should be implemented
    //in a different way.
    private KeYJavaType tempkjt = null;
    private String methodname = null;
    private Term trueLitTerm;
    private Term falseLitTerm;
    private boolean preMode = true;
    private boolean staticMode = false;
    private HashMap variable2term;
    // a dummy term to get rid of some warnings
    private Term dummy;
    // a dummy KeYJavaType to get rid of some warnings
    private KeYJavaType dt;
    private String dummyString;

    //used for parsing packagereferences
    private String packageName = "";

    private int lineOffset=0;
    private int colOffset=0;
    
    private Services services;

    public KeYJMLParser(TokenStream lexer, String filename, Services services, 
        NamespaceSet nss, TermFactory tf, TypeDeclaration cld, 
        MethodDeclaration md, JMLClassSpec cSpec, ArithOpProvider aOP){
	    this(lexer);
        setFilename(filename);
        this.cSpec = cSpec;
        this.nss = nss;
        this.services = services;
        this.tf = tf;
        this.cld = cld;
        this.md = md;
        staticMode = md != null && md.isStatic();
        this.aOP = aOP;
        variable2term = new HashMap();
        param_ns = UsefulTools.buildParamNamespace(md);
        trueLitTerm = services.getTypeConverter()
            .convertToLogicElement(BooleanLiteral.TRUE); 
        falseLitTerm = services.getTypeConverter()
            .convertToLogicElement(BooleanLiteral.FALSE); 
    }

    public KeYJMLParser(TokenStream lexer, String filename, Services services, 
        NamespaceSet nss, TermFactory tf, JMLClassSpec cSpec, 
        ArithOpProvider aOP){
	    this(lexer, filename, services, nss, tf, null, null, cSpec, aOP);
    }

    public void reportError(RecognitionException ex){
        services.getExceptionHandler().reportException(ex);
    }

    public LoopInvariant parseLoopInvariant(LoopStatement loop) 
        throws antlr.ANTLRException{
        prefix = cSpec.getInstancePrefix();
        LoopInvariant l = 
            services.getImplementation2SpecMap().getSpecForLoop(loop);
        if(l == null){
            l = new LoopInvariant(md, cld, services, loop);
        }
        param_ns = new Namespace(param_ns);
        Namespace ns = collectLocalVariables(loop, md.getBody());
        if(ns != null){
            param_ns.add(ns);
        }
        loop_spec(l);
        l.register();
        return l;
    }

    /**
     * Collects local variables of b that are visible for the statement loop.
     * Returns null if loop has not been found
     */
    private Namespace collectLocalVariables(LoopStatement loop, 
                                            StatementContainer b){
        Namespace ns = new Namespace();
        Statement s;
        for(int i = 0; i<b.getStatementCount(); i++){
            s = b.getStatementAt(i);
            if(s == loop){
                return ns;
            }
            if(s instanceof LocalVariableDeclaration){
                ArrayOfVariableSpecification vars = 
                    ((LocalVariableDeclaration) s).getVariables();
                for(int j=0; j<vars.size(); j++){
                    ns.add(vars.getVariableSpecification(j).getProgramVariable());
                }
            }
            if(s instanceof StatementContainer){
                Namespace n = collectLocalVariables(loop, 
                                                    (StatementContainer) s);
                if(n != null){ 
                    ns.add(n);
                    return ns;
                }
            }
            if(s instanceof BranchStatement){
                for(int j=0; j < ((BranchStatement) s).getBranchCount(); j++){
                    Namespace n = collectLocalVariables(loop, 
                        ((BranchStatement) s).getBranchAt(j));
                    if(n != null){ 
                        ns.add(n);
                        return ns;
                    }
                }
            }
        }
        return null;
    }

    private void addSpec(JMLMethodSpec s){
        specs = specs.add(s);
        term2old = s.getTerm2Old();
    }

    public SetOfJMLMethodSpec getSpecs(){
        return specs;
    }

    private void setPreMode(boolean mode){
        preMode = mode;
    }

    private boolean inPreMode(){
        return preMode;
    }

    private Term lookupVar(Name name) throws KeYSemanticException{
        if(variables().lookup(name)!=null){
            return tf.createVariableTerm(
                (ProgramVariable) variables().lookup(name));
        }
/*        if(nss.programVariables().lookup(name)!=null){
            return tf.createVariableTerm(
                (ProgramVariable) nss.programVariables().lookup(name));
        }*/
        if(param_ns != null && param_ns.lookup(name)!=null){
            ProgramVariable p = (ProgramVariable) param_ns.lookup(name);
            if(inPreMode()){
                return getTermForVariable(p);
            }else{
                return getOld(tf.createVariableTerm(p));
            }
        }
        return lookupAssignableVar(name);
    }

    private Term getTermForVariable(ProgramVariable p){
        if(variable2term.get(p)==null){
            Term vt = tf.createVariableTerm(p);
            variable2term.put(p,vt);
            return vt;
        }else{
            return (Term) variable2term.get(p);
        }
    }

    /** In the assignable clause no parameters may occur.
     */  
    private Term lookupAssignableVar(Name name) throws KeYSemanticException{
        ProgramVariable v = null;
        try{
            if(cSpec.lookupModelField(name)!=null){
                if(!staticMode || 
                 ((ProgramVariable) cSpec.lookupModelField(name)).isStatic()){
                    v = (ProgramVariable) cSpec.lookupModelField(name);
                    return tf.createVariableTerm(v);
                }else{
                    throw new KeYSemanticException("Nonstatic modelfield in"+
                    " static context", getLine(), getColumn(), getFilename());
                }
            }
        }catch(AmbigiousModelElementException e){
            System.err.println(e.toString());
        }
        v = getJavaInfo().getAttribute(
            name.toString(), getJavaInfo().getKeYJavaType(cld)); 
        if(v==null){
            return null;
        }
        if(v.isStatic()){
            return tf.createVariableTerm(v);
        }else if(!staticMode){
            return tf.createAttributeTerm(v, tf.createVariableTerm(
                    (ProgramVariable) cSpec.getInstancePrefix()));
        }else{
            throw new KeYSemanticException("Nonstatic field "+name+
                " in static context", getLine(), getColumn(), getFilename());
        }
    }

    private String getNameForOld(Term t){
        if(t.op() instanceof ProgramVariable){
            return ("_old"+(oldVarCount++)+"_"+ (t.toString().length() > 25 ? 
                    t.toString().substring(0,25) : t.toString())).
                    replace('.', '_').replace(':','_');
        }else{
            return ("_old"+(oldVarCount++));
        }
    }

    private Term getOld(Term t){
        Term old = (Term) term2old.get(t);
        if(old==null){
            ListOfNamed freeVariables = variables().allElements();
            if(freeVariables.size() != 0){
                Term[] args = new Term[freeVariables.size()];
                Sort[] argSorts = new Sort[freeVariables.size()];
                Sort retSort;
                if(t.sort() == Sort.FORMULA){
                    retSort = getJavaInfo().getKeYJavaType(
                        PrimitiveType.JAVA_BOOLEAN).getSort();
                }else{
                    retSort = t.sort();
                }
                int i = 0;
                IteratorOfNamed it = freeVariables.iterator();
                while(it.hasNext()){
                    ProgramVariable pv = (ProgramVariable) it.next();
                    args[i] = tf.createVariableTerm(pv);
                    argSorts[i++] = pv.sort();
                }
                Function f = new Function(new Name(getNameForOld(t)), retSort,
                                          argSorts);
                functions().add(f);
                old = tf.createFunctionTerm(f, args);
                if(isLong(t) && aOP.getCastToLong() != null){
                    old = tf.createFunctionTerm(aOP.getCastToLong(), old);
                }
            }else{
                ProgramVariable oldV;
                if(isLong(t)){
                    oldV = new ProgramVariable(
                        new ProgramElementName(getNameForOld(t)),
                        getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG));
                    old = tf.createVariableTerm(oldV);
                }else if(t.sort() == Sort.FORMULA){
                    oldV = new ProgramVariable(
                        new ProgramElementName(getNameForOld(t)),
                        getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN));
                    old = tf.createVariableTerm(oldV);
                    old = tf.createEqualityTerm(old,trueLitTerm);
                }else{
                    if(t.op() instanceof ProgramVariable){
                        oldV = new ProgramVariable(
                            new ProgramElementName(getNameForOld(t)),
                            ((ProgramVariable)t.op()).getKeYJavaType());
                    }else{
                        oldV = new ProgramVariable(
                            new ProgramElementName(getNameForOld(t)),
                            getJavaInfo().getKeYJavaType(t.sort()));
                    }
                    old = tf.createVariableTerm(oldV);
                }
                nss.programVariables().add(oldV);
            }
            term2old.put(t,old);
        }
        return old;
    }

    protected void bindVars(ListOfNamed l) {
        namespaces().setVariables(variables().extended(l));
    }

    protected void bindVar(Named n) {
        namespaces().setVariables(variables().extended(n));
    }

    protected void unbindVars() {
        namespaces().setVariables(variables().parent());
    }

    /** We have to replace arguments of functions that have been translated to
     * formulas. Usually this is done by a query, however in simple cases
     * like <code>a</code> == true which can be simplified to <code>a</code>
     * a query isn't necessary.
     * @returns a simplification of <code>arg</code> in some special cases
     * or <code>arg</code> otherwise.
     */
    protected Term simplifyArgument(Term arg){
        if(arg.op() == Op.TRUE){
            return trueLitTerm;
        }
        if(arg.op() == Op.FALSE){
            return falseLitTerm;
        }
        if(arg.op() instanceof Equality){
            if(arg.sub(0) == trueLitTerm || arg.sub(0).op() == Op.TRUE){
                return simplifyArgument(arg.sub(1));
            }
            if(arg.sub(1) == trueLitTerm || arg.sub(1).op() == Op.TRUE){
                return simplifyArgument(arg.sub(0));
            }
        }
        return arg;
    }

    private ListOfKeYJavaType getTypeList(ListOfTerm tl){
        IteratorOfTerm it = tl.iterator();
        ListOfKeYJavaType sig = SLListOfKeYJavaType.EMPTY_LIST;
        Term temp;
        while(it.hasNext()){
            temp = it.next();
            if(temp.sort() == Sort.FORMULA){
                sig = sig.append(getJavaInfo().getKeYJavaType(
                        PrimitiveType.JAVA_BOOLEAN));
            }else{
                KeYJavaType sigkjt=null;
                try{
                    sigkjt = services.getTypeConverter().
                    getKeYJavaType(temp);
                }catch(Exception e){}
                if(sigkjt == null){
                    if(!"int".equals(temp.sort().name().toString())){
                        sigkjt = getJavaInfo().getKeYJavaType(temp.sort());
                    }else{
                        sigkjt = getJavaInfo().getPrimitiveKeYJavaType(
                            temp.sort().name().toString());
                    }
                }
                sig = sig.append(sigkjt);
            }
        }
        return sig;
    }

    /** 
     * returns the translation of \\nonnullelements(t).
     */
    protected Term nonNullElements(Term t){
        Term eq1, nne, body,lRef,vTerm;
        Function f;
        eq1 = tf.createEqualityTerm(t,tf.createFunctionTerm(Op.NULL));
        eq1 = tf.createJunctorTerm(Op.NOT,new Term[]{eq1});
        if(t.sort() instanceof ArraySort && 
            ((ArraySort) t.sort()).elementSort() instanceof ObjectSort){
            ProgramVariable v = new ProgramVariable(
                new ProgramElementName("_index"+paramVarCount),
                services.getJavaInfo().getKeYJavaType("int"));
            nss.programVariables().add(v);
            LogicVariable lv = 
                new LogicVariable(new Name("index_lv"+(paramVarCount++)),
                    v.getKeYJavaType().getSort());
            vTerm = tf.createVariableTerm(v);
//            eq2 = tf.createEqualityTerm(vTerm, tf.createVariableTerm(lv));
            ProgramVariable length = services.getJavaInfo().getAttribute(
                "length", (ObjectSort) t.sort()); 
            lRef = tf.createAttributeTerm(length, t);
            f = (Function) functions().lookup(new Name("lt"));
            body = tf.createFunctionTerm(f, new Term[]{vTerm,lRef});
            f = (Function) functions().lookup(new Name("leq")); 
            body = tf.createJunctorTerm(Op.AND, new Term[]{
                    tf.createFunctionTerm(f, new Term[]{
                            toZNotation("0"),vTerm}),
                    body});
//            body = tf.createJunctorTerm(Op.IMP, new Term[]{eq2, body});
            nne = nonNullElements(tf.createArrayTerm(
                    ArrayOp.getArrayOp(t.sort()), t, vTerm));

            body = tf.createJunctorTerm(Op.IMP, new Term[]{body, nne});
            body = tf.createUpdateTerm(vTerm, tf.createVariableTerm(lv), body);
            body = tf.createQuantifierTerm(Op.ALL, lv, body);
            return tf.createJunctorTerm(Op.AND, new Term[]{eq1, body});
        }else{
            return eq1;
        }
    } 

    /**
     * creates a model method and adds it to the class specification
     */   
    protected Term createModelMethod(String name, Term result, 
        Term pre, Term post){
        ExtList el = new ExtList();
        ListOfTerm pvs = UsefulTools.getProgramVariablesFromTerm(
            result, SLListOfTerm.EMPTY_LIST);
        pvs = UsefulTools.getProgramVariablesFromTerm(pre, pvs);
        pvs = UsefulTools.getProgramVariablesFromTerm(post, pvs);
        el.add(new Public());
        if(staticMode){
            el.add(new Static());
        }else{
            pvs = 
                pvs.removeAll(tf.createVariableTerm((ProgramVariable) prefix));
        }
        IteratorOfTerm it = pvs.iterator();
        while(it.hasNext()){
            el.add(UsefulTools.getParameterDeclForVar(it.next(), services));
        }
        el.add(new ProgramElementName(name));
        KeYJavaType type = 
            services.getJavaInfo().getKeYJavaType(result.sort());
        el.add(new TypeRef(type));
        el.add(new Model());
        Term[] sub;
        MethodDeclaration meth = new MethodDeclaration(el, false);
        ProgramMethod pm = new ProgramMethod(meth, 
            services.getJavaInfo().getKeYJavaType(cld), 
            type, PositionInfo.UNDEFINED);
        cSpec.addModelMethod(pm);
        JMLNormalMethodSpec spec = new JMLNormalMethodSpec(meth, tf, 
            services, new Namespace(), new HashMap(), cSpec, nss);
        result = tf.createEqualityTerm(result, tf.createVariableTerm(
                spec.getResultVar()));
        if(pre != null){
            spec.addPre(pre);
        }
        spec.addPost(result);
        if(post != null){
            spec.addPost(post);
        }
        spec.setAssignableMode("nothing");
        services.getImplementation2SpecMap().addModifier(meth, "pure");
        services.getImplementation2SpecMap().addModifier(meth, "helper");
        return UsefulTools.createTermForQuery(pm, pvs, prefix);
    }

    protected Term buildQuantifierTerm(String q, ListOfNamed l, 
                                       Term a, Term t){
        Term body=t;
        Term eq;
        Quantifier op = null;
        if(q.equals("\\min") || q.equals("\\max")){
            boolean isLong = isLong(t);
            Function comp;
            Function minmax;
            ProgramVariable mv;
            String varName;
            if(q.equals("\\min")){
                varName = "_min";
                minmax = aOP.getMax(isLong);
                comp = (Function) functions().lookup(new Name("leq"));
            }else{
                varName = "_max";
                minmax = aOP.getMin(isLong);
                comp = (Function) functions().lookup(new Name("geq"));
            }
            mv = new ProgramVariable(
                new ProgramElementName(varName+(paramVarCount++)),
                services.getJavaInfo().getKeYJavaType(
                    isLong ? "long" : "int"));
            nss.programVariables().add(mv);
            Term mvTerm = tf.createVariableTerm(mv);
            Term mTerm = tf.createFunctionTerm(minmax);
            Term ex1 = buildQuantifierTerm("\\exists", l, null, a);
            ex1 = tf.createJunctorTerm(Op.NOT, new Term[]{ex1});
            ex1 = tf.createJunctorTerm(Op.AND, new Term[]{ex1,
                tf.createEqualityTerm(mTerm, mvTerm)});
            Term ex2 = buildQuantifierTerm("\\exists", l, a, 
                tf.createEqualityTerm(mvTerm, t));
            Term all1 = buildQuantifierTerm("\\forall", l, a, 
                tf.createFunctionTerm(comp, new Term[]{mvTerm, t}));
            all1 = tf.createJunctorTerm(Op.AND, new Term[]{all1, ex2});
            all1 = tf.createJunctorTerm(Op.OR, new Term[]{ex1, all1});
            return createModelMethod(q.substring(1)+(mmCount++), mvTerm, 
                null, all1);
        }
        if(l==SLListOfNamed.EMPTY_LIST){
            if(a!=null){
                if(q.equals("\\exists")){
                    body=tf.createJunctorTerm(Op.AND,new Term[]{a, body});
                }else{
                    body=tf.createJunctorTerm(Op.IMP,new Term[]{a, body});
                }
            }
            return body;
        }else{
            if(q.equals("\\forall")){
                op = Op.ALL;    
            }else if(q.equals("\\exists")){
                op = Op.EX;
            }else{
                throw new IllegalArgumentException("Quantifier "+q+
                    " not supported");
            }
            body = buildQuantifierTerm(q, l.tail(), a, t);
            ProgramVariable v = (ProgramVariable) l.head();
            LogicVariable lv = new LogicVariable(new Name(v.name()+"_jml_lv"+
                    (paramVarCount++)),
                v.getKeYJavaType().getSort());
            Term vTerm = tf.createVariableTerm(v);
            // if the quantified variables have an objecttype, we only quantify
            // over objects that are already created (nevertheless the 
            // quantified variable may be null) 
            if(v.getKeYJavaType().getSort() instanceof ObjectSort){
                Term cnCond = null;
                ProgramVariable created = 
                    services.getJavaInfo().getAttribute("<created>",
                        (ObjectSort) v.getKeYJavaType().getSort());
                Term createdTerm = tf.createAttributeTerm(created, vTerm);
                createdTerm = tf.createEqualityTerm(createdTerm, trueLitTerm);
                Term nullComp = tf.createEqualityTerm(vTerm, 
                    tf.createFunctionTerm(Op.NULL));
                cnCond = tf.createJunctorTerm(Op.AND, new Term[]{createdTerm,
                    tf.createJunctorTerm(Op.NOT, new Term[]{nullComp})});
                cnCond = 
                    tf.createJunctorTerm(Op.OR, new Term[]{cnCond, nullComp});
                if(op == Op.ALL){
                    body = tf.createJunctorTerm(Op.IMP, new Term[]{cnCond,body});
                }else{
                    body = tf.createJunctorTerm(Op.AND, new Term[]{cnCond,body});
                }
            }
            return UsefulTools.createQuantifierTermWithPV(op, vTerm, lv, body);
        }  
    }

    private boolean isLong(Term t){
        if(t.sort() == Sort.FORMULA) return false;
        if(t.op() instanceof ProgramVariable && 
            ((ProgramVariable)t.op()).getKeYJavaType() != null){
            return PrimitiveType.JAVA_LONG.equals(((ProgramVariable)t.op()).
                getKeYJavaType().getJavaType());
        }
        if(t.op().name().toString().startsWith("castTo")){
            return t.op().name().toString().equals("castToLong");
        }
        for(int i=0; i<t.arity(); i++){
            if(isLong(t.sub(i))){
                return true;
            }
        }
        return false;
    }

	public JavaInfo getJavaInfo() {
	    return getServices().getJavaInfo();
    }

	public Services getServices() {
	    return services;
    }

    public NamespaceSet namespaces(){
        return nss;
    }

    public Namespace sorts() {
        return namespaces().sorts();
    }

    public Namespace variables() {
        return namespaces().variables();
    }
    
    public Namespace functions() {
        return namespaces().functions();
    }

    public Namespace choices(){
        return namespaces().choices();
    }

    public void parseClassSpec() throws antlr.ANTLRException {
        top();
    }

    public void parseMethodDecl(String comments) throws antlr.ANTLRException {
        methoddecl2(comments);        
    }

    public void parseFieldDecl() throws antlr.ANTLRException {
        field();        
    }

    public Term toZNotation(String s){
        Term[] t1=new Term[1];
        Operator v;
        Term a;
        v=(Function) functions().lookup(new Name("#"));
        t1[0] = tf.createFunctionTerm((Function)v);
        v = (Function) functions().lookup(new Name(s.substring(0,1)));
        a = tf.createFunctionTerm((Function)v,t1);
        for(int i=1;i<s.length();i++){
            t1[0]=a;    
            v = (Function)functions().lookup(new Name(s.substring(i,i+1)));
            a = tf.createFunctionTerm((Function)v,t1);
        }   
        t1[0]=a;
        v = (Function) functions().lookup(new Name("Z"));
        a = tf.createFunctionTerm((Function)v,t1);
        return a;
    }

    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine()+lineOffset;
        } catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }

    protected int getColumn() {
        int 	col = -1;
        try {
            col = LT(0).getColumn()+colOffset;
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }   

}

top : 
        // parses the pure modifier in cases in which it refers to a type declaration 
        ("pure") => "pure"
        {
            services.getImplementation2SpecMap().setPure(
                getJavaInfo().getKeYJavaType(cld), nss);
        }
    |   
        (field)*
    ;

loop_spec [LoopInvariant l]:
        (   
            loop_invariant[l]
        |
            assignableclause[l]
        |
            variant_function[l]
        )+
    ;

loop_invariant[LoopInvariant l]
{
    Term t;
}:
        maintaining_keyword t=predicate ";"
        {
            l.addInvariant(t);
        }
;

maintaining_keyword :
        "maintaining" 
    | 
        "maintaining_redundantly"
    | 
        "loop_invariant" 
    | 
        "loop_invariant_redundantly"
;

variant_function[LoopInvariant l] 
{
    Term t;
}: 
        decreasing_keyword t=specexpression ";"
        {
            l.setVariant(t);
        }
    ;

decreasing_keyword : 
        "decreasing" 
    | 
        "decreasing_redundantly"
    | 
        "decreases" 
    | 
        "decreases_redundantly"
    ;



field 
{
    SetOfString mods=SetAsListOfString.EMPTY_SET;
}: 
        {term2old = cSpec.getTerm2Old(); prefix = cSpec.getInstancePrefix();}
        (
            (("model")? "import") => import_definition
        |
            (
                mods=modifiers
                (
                    jmldeclaration[mods]
                |
                    memberdecl[mods]
                )
            )
        )
    ;

import_definition :
        ("model")? "import" name_star ";"
    ;

name_star :
        IDENT (DOT (IDENT | MULT))*
    ;

memberdecl[SetOfString mods] :
        variabledecls[mods]
    ;

//not used due to the necessity that the methoddeclaration has to be parsed
//before the method specification
//methoddecl :
//        methodspecification 
//        methoddecl2[""]
//        {
//            addMethod2Specs();
//        }
//;

methoddecl2[String comments]
{
    KeYJavaType type;
    SetOfString mods=SetAsListOfString.EMPTY_SET;
}:
        mods=modifiers type=typespec methodhead[type, mods, comments] ";"
    ;

methodhead[KeYJavaType type, SetOfString mods, String comments]
{
    LinkedList l = new LinkedList();
}:
        id:IDENT "(" (l = paramdeclarationlist)? ")"
        {
            ExtList el = new ExtList();
            if(comments != null){
                el.add(new Comment(comments));
            }
            el.add(new Public());
            if(mods.contains("static")){
                el.add(new Static());
            }
            if(mods.contains("private")){
                el.add(new Private());
            }else if(mods.contains("public")){
                el.add(new Public());
            }else if(mods.contains("protected")){
                el.add(new Protected());
            }
            el.add(new Model());
            Object[] oarr = l.toArray();  
            for(int i=0; i<oarr.length; i++){
                el.add((ParameterDeclaration) oarr[i]);
            }
            el.add(new ProgramElementName(id.getText()));
            el.add(new TypeRef(type));
            md = new MethodDeclaration(el, false);
            if(mods.contains("pure")){
                services.getImplementation2SpecMap().addModifier(md, "pure");
            }
            if(mods.contains("helper")){
                if(!md.isPrivate()){
                    throw new KeYSemanticException("Helper methods must be "+
                        "private but "+md.getName()+" isn't", 
                        getLine(), getColumn(), getFilename());
                }
                services.getImplementation2SpecMap().addModifier(md, "helper");
            }
            cSpec.addModelMethod(new ProgramMethod(md, 
                    services.getJavaInfo().getKeYJavaType(cld), 
                    type, PositionInfo.UNDEFINED));
        }
;

paramdeclarationlist returns[LinkedList pl = new LinkedList()]
{
    ParameterDeclaration pd = null;
}:
        pd = paramdeclaration {pl.add(pd);} 
        ("," pd = paramdeclaration {pl.add(pd);})*
;

paramdeclaration returns[ParameterDeclaration pd = null]
{
    int dim = 0;
    KeYJavaType type;
}:
        type=typespec id:IDENT (dim=dims)?
        {
            ProgramVariable v = new ProgramVariable(new ProgramElementName(
                id.getText()), type); 
            param_ns.add(v);
            pd = new ParameterDeclaration(new Modifier[0], 
                new TypeRef(type), new VariableSpecification(v), false);
        }
;

variabledecls[SetOfString mods]
{
    KeYJavaType type;
}:
        type=typespec variabledeclarators[type, mods] ";"
        (jml_datagroup_clause)?
    ;

//datagroups
//not implemented --->
jml_datagroup_clause :
        in_group_clause | maps_into_clause
    ;

in_group_clause :
        in_keyword group_list ";" 
    ;

in_keyword :
        "in" | "in_redundantly"
    ;

group_list :
        group_name ("," group_name)*
    ;

group_name :
        group_name_prefix IDENT
    ;

group_name_prefix :
        "this" DOT | "super" DOT
    ;

maps_into_clause :
        maps_keyword member_field_ref INTO group_list ";"
    ;

maps_keyword :
        "maps" | "maps_redundantly"
;

member_field_ref :
        (IDENT ".") => IDENT "." maps_member_ref_expr
    | 
        maps_array_ref_expr ( "." maps_member_ref_expr )?
    ;

maps_member_ref_expr :
        IDENT | MULT 
    ;

maps_array_ref_expr :
        IDENT maps_spec_array_dim ( maps_spec_array_dim )* 
;

maps_spec_array_dim :
        "[" dummy = specarrayrefexpr "]"
    ;

//   <---   not implemented

variabledeclarators[KeYJavaType type, SetOfString mods]:
        variabledeclarator[type, mods] ("," variabledeclarator[type, mods])*
    ;

variabledeclarator[KeYJavaType type, SetOfString mods]
{
    int dim = 0;
}:
        id:IDENT (dim=dims)?
        {
            type = services.getJavaInfo().getKeYJavaType(
                ArraySortImpl.getArraySortForDim(type.getSort(), dim, 
                    services.getJavaInfo().getObjectSort(),
                    services.getJavaInfo().getCloneableSort()));
            cSpec.addModelVariable(new ProgramVariable(
                    new ProgramElementName(id.getText()),type, 
                    getJavaInfo().getKeYJavaType(cld),
                    mods != null && mods.contains("private"),
                    mods != null && mods.contains("static"),
                    true));
        }
    ;

modifiers returns[SetOfString mods=SetAsListOfString.EMPTY_SET]
{
    String m;
}:
        (m=modifier {mods = mods.add(m);})*
    ;

modifier returns[String s=null]:
        s=privacy
    |
        "instance" {s="instance"; prefix = cSpec.getInstancePrefix();}
    |
        "static" {s="static";}
    |
        "model" {s = "model";}
    |
        "pure" {s = "pure";}
    |
        "helper" {s = "helper";}
    | 
        "ghost" 
    | 
        "uninitialized"
    | 
        "spec_java_math" 
    | 
        "spec_safe_math" 
    | 
        "spec_bigint_math"
    | 
        "code_java_math" 
    | 
        "code_safe_math" 
    | 
        "code_bigint_math"
    | 
        "non_null"
    ;

privacy returns[String s=null]:
        "public" {s="public";}
    |
        "protected" {s="protected";}
    |       
        "private" {s="private";}
    |
        "spec_public" 
    | 
        "spec_protected"
    ;

jmldeclaration[SetOfString mods] : 
        invariant[mods]
    |
        historyconstraint[mods]
    |
        representsdecl[mods]
    | 
        initially_clause
    | 
        monitors_for_clause
    | 
        readable_if_clause
    | 
        writable_if_clause
    | 
        "axiom" dummy = predicate ";"
    ;

//not implemented --->
initially_clause :
        "initially" dummy = predicate ";"
    ;

readable_if_clause :
        "readable" IDENT "if" dummy = predicate ";"
    ;

writable_if_clause :
        "writable" IDENT "if" dummy = predicate ";"
    ;

monitors_for_clause : 
        "monitors_for" IDENT
        l_arrow_or_eq spec_expression_list ";"
    ;
//  <--- not implemented

representsdecl[SetOfString mods]
{
    Term t1,t2;
}:
        representskeyword t1=storerefexpression
        (
            l_arrow_or_eq t2=specexpression
            {
                cSpec.addRepresents(t1, t2);
            }
        |
            SUCH_THAT t2=predicate
            {
                cSpec.addSuchThat(t1, t2);
            }
        )
        ";"
    ;

representskeyword:
        "represents"
    |
        "represents_redundantly"
    ;

l_arrow_or_eq:
        "<-" 
    |
        "="
    ;

invariant[SetOfString mods]{
    Term t=null;
} :
        invariantkeyword t=predicate ";"
        {
            if(mods.contains("instance")){
                cSpec.addInstanceInvariant(t);
            }else if(mods.contains("static") || 
                     cld instanceof InterfaceDeclaration){
                cSpec.addStaticInvariant(t);                
            }else{
                cSpec.addInstanceInvariant(t); 
            }
        }
    ;

invariantkeyword :
        "invariant" 
    |
        "invariant_redundantly" 
    ;

historyconstraint[SetOfString mods]{
    Term t=null;
} :
        constraintkeyword t=predicate ";"
        {
            if(mods.contains("static")){
                cSpec.addStaticConstraint(t);
            }else{
                cSpec.addInstanceConstraint(t);
            }
        }
    ;

constraintkeyword :
        "constraint" 
    |
        "constraint_redundantly"
    ;

methodspecification : 
        specification
    |
        //HACK(?)
        (pure_helper)*
    |
        extendingspecification
    ;

pure_helper :
       (
            "pure" 
            {
                if(!services.getImplementation2SpecMap().
                    getModifiers(md).contains("pure")){
                    new JMLPuritySpec(md, tf, services, param_ns, 
                        term2old, cSpec, nss);
                    services.getImplementation2SpecMap().
                        addModifier(md, "pure");
                }
            }
        |
            "helper" 
            {
                if(!md.isPrivate()){
                    throw new KeYSemanticException("Helper methods must be "+
                        "private but "+md.getName()+" isn't", 
                        getLine(), getColumn(), getFilename());
                }
                services.getImplementation2SpecMap().addModifier(md, "helper");
            }
        )
    ;

extendingspecification :
        "also" specification
    ;

specification :
        speccaseseq
    ;

speccaseseq :
        speccase (ALSO speccase)*
    ;

speccase :
        ((dummyString=privacy)? "model_program") => model_program
    |
        lightweightspeccase
    |
        heavyweightspeccase
    |
        code_contract_spec
    ;

//not implemented --->
code_contract_spec : 
        "code_contract" (code_contract_clause)+
    ;

code_contract_clause :
        accessible_clause
    | 
        callable_clause
    | 
        measured_clause 
    ;

measured_clause :
        measured_by_keyword
        (
            NOT_SPECIFIED ";"
        |
            dummy=specexpression ( "if" dummy=predicate )? ";"
        )
    ;

measured_by_keyword :
        "measured_by" 
    |
        "measured_by_redundantly"
    ;

accessible_clause 
{
    SetOfTerm sot;
}:
        accessible_keyword sot=conditionalstorereflist[null] ";"
    ;

accessible_keyword :
        "accessible" 
    |
        "accessible_redundantly"
    ;

callable_clause :
        callable_keyword callable_methods_list ";"
    ;

callable_keyword :
        "callable" 
    |
        "callable_redundantly"
    ;

callable_methods_list :
        method_name_list | dummyString=storerefkeyword
    ;

method_name_list :
        method_name ( "," method_name )*
    ;

method_name :
        method_ref ( "(" ( param_disambig_list )? ")" )?
    ;

method_ref :
        method_ref_start ( DOT method_ref_rest )*
    | 
        "new" dt=referencetype
    ;

method_ref_start :
        "super" | "this" | IDENT | OTHER
    ;

method_ref_rest :
        "this" | IDENT | OTHER
    ;

param_disambig_list :
        param_disambig ( "," param_disambig )*
    ;

param_disambig 
{
    int d;
}:
        dt=typespec ( IDENT ( d=dims )? )?
    ;

model_program : 
        ( dummyString=privacy )? 
        {
            throw new NotSupportedExpressionException("Model Program");
        }
        "model_program" LBRACE
    ;

// <--- not implemented

lightweightspeccase 
{
    JMLMethodSpec s=new JMLMethodSpec(md, tf, services, 
        param_ns, term2old, cSpec, nss);
}:
        genericspeccase[s]
    ;

heavyweightspeccase 
{
    String p = null;
}:
        (p=privacy)?  
        (
            "normal_behavior" 
            normalspeccase[new JMLNormalMethodSpec(md, tf, services, 
                    param_ns, term2old, cSpec, nss)]
        |
            "behavior" genericspeccase[new JMLMethodSpec(md, tf, 
                    services, param_ns, term2old, cSpec, nss)]
        |
            "exceptional_behavior" 
            exceptionalspeccase[new JMLExceptionalMethodSpec(md, tf, 
                    services, param_ns, term2old, cSpec, nss)]
        )
    ;

genericspeccase [JMLMethodSpec s] 
{
    result = s.getResultVar();
    prefix = s.getPrefix();
    addSpec(s);
}:
        specheader[s] (genericspecbody[s])? 
    |
        genericspecbody[s] 
     ;

normalspeccase [JMLNormalMethodSpec s]
{
    result = s.getResultVar();
    prefix = s.getPrefix();
    addSpec(s);
} :
        specheader[s] (normalspecbody[s])?
    |
        normalspecbody[s]
    ;

exceptionalspeccase [JMLExceptionalMethodSpec s] 
{
    result = s.getResultVar();
    prefix = s.getPrefix();
    addSpec(s);
} :
        specheader[s] (exceptionalspecbody[s])?
    |
        exceptionalspecbody[s]
    ;


specheader[JMLMethodSpec s] :
        requiresclause[s] (requiresclause[s])*
    ;
        
genericspecbody[JMLMethodSpec s] :
        simplespecbodyclause[s] (simplespecbodyclause[s])*
    |
        LBRACE "|" genericspeccaseseq[s] "|" RBRACE
    ;

genericspeccaseseq[JMLMethodSpec s] 
{
    JMLMethodSpec copy = s.copy();
    services.getImplementation2SpecMap().removeMethodSpec(copy);
}:
        genericspeccase[s] ( "also" genericspeccase[copy.copy()] )*
;

normalspecbody[JMLNormalMethodSpec s] :
        normalspecclause[s] (normalspecclause[s])*
    |
        LBRACE "|" normalspeccaseseq[s] "|" RBRACE
    ;

normalspeccaseseq[JMLNormalMethodSpec s] 
{
    JMLNormalMethodSpec copy = (JMLNormalMethodSpec) s.copy();
    services.getImplementation2SpecMap().removeMethodSpec(copy);
}:
        normalspeccase[s] 
        ( 
            "also" normalspeccase[(JMLNormalMethodSpec) copy.copy()] 
        )*
;

exceptionalspecbody[JMLExceptionalMethodSpec s] :
        exceptionalspecclause[s] (exceptionalspecclause[s])*
    |
        LBRACE "|" exceptionalspeccaseseq[s] "|" RBRACE
    ;

exceptionalspeccaseseq[JMLExceptionalMethodSpec s] 
{
    JMLExceptionalMethodSpec copy = (JMLExceptionalMethodSpec) s.copy();
    services.getImplementation2SpecMap().removeMethodSpec(copy);
}:
        exceptionalspeccase[s] 
        ( 
            "also" exceptionalspeccase[(JMLExceptionalMethodSpec) copy.copy()] 
        )*
;

normalspecclause[JMLMethodSpec s] :
        ensuresclause[s]
    |
        assignableclause[s]
    |
        divergesclause[s]
    |
        whenclause[s]
    |
        durationclause[s]
    |
        workingspaceclause[s]
    ;

simplespecbodyclause[JMLMethodSpec s] :
        normalspecclause[s]
    |
        signalsclause[s]
    ;

exceptionalspecclause[JMLMethodSpec s] :
        assignableclause[s]
    |
        signalsclause[s]
    |
        divergesclause[s]
    |
        whenclause[s]
    |
        durationclause[s]
    |
        workingspaceclause[s]
    ;

//not implemented
whenclause[JMLMethodSpec s] :
        whenkeyword dummy = predornot ";"
    ;

whenkeyword:
        "when" 
    | 
        "when_redundantly"
    ;


//not implemented
durationclause[JMLMethodSpec s] :
        durationkeyword dummy = predornot ";"
    ;

durationkeyword:
        "duration" 
    | 
        "duration_redundantly"
    ;

//not implemented
workingspaceclause[JMLMethodSpec s] :
        workingspacekeyword 
        (
            NOT_SPECIFIED
        |
            dummy = specexpression (IF dummy = predicate)?
        )
        ";"
    ;

workingspacekeyword:
        "working_space" 
    | 
        "working_space_redundantly"
    ;

divergesclause[JMLMethodSpec s]
{
    Term t=null;
}:
        divergeskeyword 
        {
            setPreMode(true);
        } 
        t=predornot 
        {
            s.addDiverges(t);
        }
        ";"
; 

divergeskeyword :
        "diverges"
    |
        "diverges_redundantly"
    ;

assignableclause[AssignableSpec s] 
{
    SetOfTerm cs = null;
}:
        assignablekeyword cs = conditionalstorereflist[s] ";"
        {
            s.addAssignable(cs);
        }
;

assignablekeyword :
        "assignable"
    |
        "modifies"
    |
        "modifiable"
    | 
        "assignable_redundantly"
    |   
        "modifiable_redundantly"
    | 
        "modifies_redundantly"
;

conditionalstorereflist[AssignableSpec s] 
    returns [SetOfTerm set = SetAsListOfTerm.EMPTY_SET]
{
    Term t=null;
} :
        t=conditionalstoreref[s]
        {
            if(t != null){
                set = set.add(t);
            }
        }
        (
            "," t=conditionalstoreref[s]
            {
                if(t != null){
                    set = set.add(t);
                }
            }
        )* 
;

conditionalstoreref[AssignableSpec s] returns [Term t=null]
{
    Term t1=null;
}:
        t=storeref[s] ("if" t1=predicate)? 
;

// not implemented
storereflist :
        dummy=storeref[null] ("," dummy=storeref[null])*
;

storeref[AssignableSpec spec] returns [Term t=null]
{
    String s;
}:
        t=storerefexpression
    |
        s=storerefkeyword 
        {
            if(spec != null){
                spec.setAssignableMode(s);
            }
        }
    |
        INFORMAL_DESCRIPTION
;

storerefkeyword returns [String s=null]:
        (
            NOTHING {s="nothing";}
        |
            EVERYTHING {s="everything";}
        |
            NOT_SPECIFIED {s="not_specified";}
        |
            PRIVATEDATA {s="private_data";}
        )
;

storerefexpression returns [Term t=null]
{
    String s;
    KeYJavaType kjt = null;
}:
        s=storerefname 
        {
            if(s.equals("this")){
                if(staticMode){
                    throw new KeYSemanticException("this in static context", 
                        getLine(), getColumn(), getFilename());
                }
                t = tf.createVariableTerm((ProgramVariable) prefix);
            }else{
                t=lookupVar(new Name(s));
            }
            if(t == null){
                kjt = getJavaInfo().getTypeByClassName(s);
            }   
        }
        (
            t=storerefnamesuffix[t, kjt] 
        )*
    ;

storerefnamesuffix[Term localPrefix, KeYJavaType kjt] returns [Term t=null]
{
    String s;
    Term index = null;
}:
        (
            (DOT "this") => DOT "this"
            {
                t = tf.createVariableTerm((ProgramVariable) prefix);
            }
        |
            DOT id:IDENT 
            {
                s = id.getText();
                if(localPrefix != null){
                    kjt = services.getTypeConverter().getKeYJavaType(
                        localPrefix);
                }
                //kjt can be null because of the not implemented parts of
                //specarrayrefexpr
                if(kjt != null){
                    ProgramVariable v = (ProgramVariable) getJavaInfo().
                        getAttribute(s, kjt);
                    if ( v == null) {
                        throw new NotDeclException("Attribute ", s, 
                            getFilename(), getLine(), getColumn());
                    }
                    if(v.isStatic()){
                        t = tf.createVariableTerm(v);
                    }else{
                        t = tf.createAttributeTerm(v,localPrefix);
                    }
                }
            }
        |
            "[" index = specarrayrefexpr "]"
            {
                if(index != null && localPrefix != null){
                    t = tf.createArrayTerm(ArrayOp.
                    	getArrayOp(localPrefix.sort()), localPrefix, index);
                }
            }
        )
    ;

specarrayrefexpr returns [Term t = null] :
        t=specexpression 
        (/*not implemented*/ DOTDOT t = specexpression 
            {throw new NotSupportedExpressionException("JML construct .. in "+
                "array indices");})?
    |
        //not implemented
        MULT {throw new NotSupportedExpressionException("JML construct .. in "+
                "array indices");}
    ;

storerefname returns [String s=null] :
        id:IDENT {s=id.getText();}
    |   
        "this" 
        {
            s = "this";
        }
    |
        "super"
        {
            s = "super";
        }
;

signalsclause[JMLMethodSpec s]{
    KeYJavaType exc=null;
    Term t=null;
    ProgramVariable v = null;
}:
        signalskeyword {setPreMode(false);} 
        "(" exc=referencetype 
        (id:IDENT 
            {
                v = new ProgramVariable(
                    new ProgramElementName(id.getText()), exc);
                bindVar(v);
            }
        )? 
        ")" t=predornot ";"
        {
            if(t != null){
                s.addSignals(exc, v, t);
            }
            if(v != null){
                unbindVars();
            }
        }
;

signalskeyword:
        (   
            "signals"
        |
            "exsures"
        |
            "signals_redundantly"
        |
            "exsures_redundantly"
        )
;

requiresclause[JMLMethodSpec s] 
{
    Term t=null; 
}:
        requireskeyword {setPreMode(true);} t=predornot 
        {
            if(t!=null) s.addPre(t); 
        } 
        ";"
    ;

requireskeyword :
        (
            "requires"
        |
            "pre"
        |
            "requires_redundantly"
        |
            "pre_redundantly"
        )
    ;

ensuresclause[JMLMethodSpec s] 
{
    Term t=null; 
}:
        ensureskeyword {setPreMode(false);} t=predornot 
        {
            if(t!=null) s.addPost(t); 
        } 
        ";"
    ;

ensureskeyword :
        (
            "ensures"
        |
            "post"
        |
            "ensures_redundantly"
        |
            "post_redundantly"
        )
    ;

predornot returns [Term t=null]:
        t = predicate 
    |
        NOT_SPECIFIED
    ;

predicate returns [Term t=null]:
        t = specexpression
    ;

specexpression returns [Term t=null]:
        t = expression
    ;

spec_expression_list :
        dummy = specexpression ("," dummy = specexpression)*
    ;

expression returns [Term t=null]:
        t=assignmentexpr
    ;

assignmentexpr returns [Term t=null]:
        t=conditionalexpr
    ;

conditionalexpr returns [Term t=null]
{
    Term t1,t2;
}:
        t=equivalenceexpr 
        (
            "?" t1=conditionalexpr ":" t2=conditionalexpr
            {
                if(t1.sort() == Sort.FORMULA){
                    t1 = tf.createJunctorTerm(Op.IMP,new Term[]{t,t1});
                    t2 = tf.createJunctorTerm(Op.IMP,new Term[]{
                            tf.createJunctorTerm(Op.NOT,new Term[]{t}),t2});
                    t = tf.createJunctorTerm(Op.AND,new Term[]{t1,t2});
                }else{
                    KeYJavaType type;
                    if(isLong(t1) || isLong(t2)){
                        type = getJavaInfo().getKeYJavaType(
                            PrimitiveType.JAVA_LONG);
                    }else if(t1.sort() instanceof ObjectSort){
                        type = getJavaInfo().getObjectKeYJavaType();
                    }else{
                        type = getJavaInfo().getKeYJavaType(t1.sort());
                    }
                    ProgramVariable v = new ProgramVariable(
                        new ProgramElementName(
                            "_cond_expre_var"+(paramVarCount++)), type);
                    nss.programVariables().add(v);
                    Term vt = tf.createVariableTerm(v);
                    t1 = tf.createJunctorTerm(Op.IMP,
                        new Term[]{t, tf.createEqualityTerm(vt,t1)});
                    t2 = tf.createJunctorTerm(Op.IMP,
                        new Term[]{tf.createJunctorTerm(Op.NOT,new Term[]{t}),
                            tf.createEqualityTerm(vt,t2)});
                    t = createModelMethod("condExpr"+(mmCount++), vt, null, 
                        tf.createJunctorTerm(Op.AND,new Term[]{t1,t2}));
                }
            }
        )?
    ;

equivalenceexpr returns [Term t=null]
{
    Term a=null;
}:
        t=impliesexpr 
        (
            EQV a=equivalenceexpr 
            {
                t = tf.createJunctorTerm(Op.EQV,new Term[]{t,a});
            }
        |
            ANTIV a=equivalenceexpr 
            {
                t = tf.createJunctorTerm(Op.EQV,new Term[]{t,a});
                t = tf.createJunctorTerm(Op.NOT,new Term[]{t});
            }
        )?
    ;
        
impliesexpr returns [Term t=null]
{
    Term a=null;
}:
        t=logicalorexpr 
        (
            "==>" a=impliesnonbackwardexpr
            { 
                t = tf.createJunctorTerm(Op.IMP,new Term[]{t,a});
            }
        |
            (
                "<==" a=logicalorexpr
                { 
                    t = tf.createJunctorTerm(Op.IMP,new Term[]{a,t});
                }
            )+
        )?
;

impliesnonbackwardexpr returns [Term t=null]
{
    Term a=null;
}:
        t=logicalorexpr ("==>" a=impliesnonbackwardexpr
            { t = tf.createJunctorTerm(Op.IMP,new Term[]{t,a});})?
;        

logicalorexpr returns [Term t=null]
{
    Term a=null;
}:
        t=logicalandexpr ("||" a=logicalorexpr 
            {t = tf.createJunctorTerm(Op.OR, new Term[]{t,a});})?
;

logicalandexpr returns [Term t=null]
{
    Term a=null;
}:
        t=inclusiveorexpr ("&&" a=logicalandexpr
            {t = tf.createJunctorTerm(Op.AND, new Term[]{t,a});})?
;

inclusiveorexpr returns [Term t=null]
{
    Term a=null;
}:
        t=exclusiveorexpr 
        (
            "|" a=inclusiveorexpr
            {
                if(t.sort() == Sort.FORMULA && a.sort() == Sort.FORMULA){
                    t = tf.createJunctorTerm(Op.OR, new Term[]{t,a});
                }else if(t.sort().equals(a.sort())){
                    Function f = aOP.getOr(isLong(t) || isLong(a));
                    t = tf.createFunctionTerm(f,new Term[]{t,a});
                }else{
                    throw new antlr.SemanticException(
                        "Bitwise-Or not valid with subterms "+t+" and "+a);
                }
            }
        )?
;

exclusiveorexpr returns [Term t=null]
{
    Term a,na,nt,a1,a2;
}:
        t=andexpr 
        (XOR a=exclusiveorexpr 
            {
                if(t.sort() == Sort.FORMULA && a.sort() == Sort.FORMULA){
                    na = tf.createJunctorTerm(Op.NOT,new Term[]{a});
                    nt = tf.createJunctorTerm(Op.NOT,new Term[]{t});
                    a1 = tf.createJunctorTerm(Op.AND, new Term[]{t,na});
                    a2 = tf.createJunctorTerm(Op.AND, new Term[]{nt,a});    
                    t = tf.createJunctorTerm(Op.OR, new Term[]{a1,a2});
                }else if(t.sort().equals(a.sort())){
                    Function f = aOP.getXor(isLong(t) || isLong(a));
                    t = tf.createFunctionTerm(f,new Term[]{t,a});
                }else{
                    throw new antlr.SemanticException(
                        "Bitwise-Xor not valid with subterms "+t+" and "+a);
                }
            }
        )?
;

andexpr returns [Term t=null]
{
    Term a=null;
}:
        t=equalityexpr
        (
            "&" a=andexpr
            {
                if(t.sort() == Sort.FORMULA && a.sort() == Sort.FORMULA){
                    t = tf.createJunctorTerm(Op.AND, new Term[]{t,a});
                }else if(t.sort().equals(a.sort())){
                    Function f = aOP.getAnd(isLong(t) || isLong(a));
                    t = tf.createFunctionTerm(f,new Term[]{t,a});
                }else{
                    throw new antlr.SemanticException(
                        "Bitwise-And not valid with subterms "+t+" and "+a);
                }
            }
        )?
;

equalityexpr returns [Term t=null]
{
    Term a1,a2;
}:
        t=relationalexpr 
        (
            "==" a1=equalityexpr {t=tf.createEqualityTerm(t,a1);}
        |
            "!=" a1=equalityexpr 
            {
                a2=tf.createEqualityTerm(t,a1);
                t=tf.createJunctorTerm(Op.NOT,new Term[]{a2});
            }
        )?
;

relationalexpr returns [Term t=null]
{
    Term a1;
    Function f;
    KeYJavaType type;
}:
        t=shiftexpr
        (
            "<" a1=shiftexpr 
            {
                f = (Function) functions().lookup(new Name("lt"));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        |
            ">" a1=shiftexpr
            {
                f = (Function) functions().lookup(new Name("gt"));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        |
            "<=" a1=shiftexpr
            {
                f = (Function) functions().lookup(new Name("leq"));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        |
            ">=" a1=shiftexpr
            {
                f = (Function) functions().lookup(new Name("geq"));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        |
            "instanceof" type=typespec 
            {
                SortDefiningSymbols s = (SortDefiningSymbols)(type.getSort());
                InstanceofSymbol func
                    = (InstanceofSymbol)s.lookupSymbol(InstanceofSymbol.NAME);
                t = tf.createFunctionTerm(func, t);
                t = tf.createEqualityTerm(t,trueLitTerm);
            }
        )?
;

shiftexpr returns [Term t=null]
{
    Term a=null;
}:
        t=additiveexpr
        (
            ">>" a=additiveexpr 
            {
                Function f = aOP.getShiftRight(isLong(t));
                t = tf.createFunctionTerm(f,new Term[]{t,a});
            }
        |   
            "<<"
            {
                Function f = aOP.getShiftLeft(isLong(t));
                t = tf.createFunctionTerm(f,new Term[]{t,a});
            }
        |   
            ">>>"
            {
                Function f = aOP.getUnsignedShiftRight(isLong(t));
                t = tf.createFunctionTerm(f,new Term[]{t,a});
            }
        )*
;

additiveexpr returns [Term t=null]
{
    Term a1;
    Function f;
}:
        t=multexpr
        (
            "+" a1=additiveexpr
            {
                f = aOP.getAdd(isLong(t) || isLong(a1));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        |
            "-" a1=additiveexpr
            {
                f = aOP.getSub(isLong(t) || isLong(a1));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        )?
;

multexpr returns [Term t=null]
{
    Term a1;
    Function f;
}:
        t=unaryexpr
        (
            MULT a1=multexpr
            {  
                f = aOP.getMul(isLong(t) || isLong(a1));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        |
            DIV a1=multexpr
            { 
                f = aOP.getDiv(isLong(t) || isLong(a1));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        |
            "%" a1=multexpr
            { 
                f = aOP.getMod(isLong(t) || isLong(a1));
                t = tf.createFunctionTerm(f,new Term[]{t,a1});
            }
        )?
;

unaryexpr returns [Term t=null]{
    Function neg;
    KeYJavaType type;
}:
        "+" t=unaryexpr
    |
        "-" t=unaryexpr 
        {
            neg = aOP.getMinus(isLong(t));
            t = tf.createFunctionTerm(neg,new Term[]{t});
        }
    |
        ("(" builtintype ")") =>
        "(" type=builtintype ")" t=unaryexpr
        {
            if(PrimitiveType.JAVA_BYTE.equals(type.getJavaType())){
                if(aOP.getCastToByte() != null){
                    t = tf.createFunctionTerm(aOP.getCastToByte(),t); 
                }
            }else if(PrimitiveType.JAVA_SHORT.equals(type.getJavaType())){
                if(aOP.getCastToShort() != null){
                    t = tf.createFunctionTerm(aOP.getCastToShort(),t); 
                }
            }else if(PrimitiveType.JAVA_INT.equals(type.getJavaType())){
                if(aOP.getCastToInt() != null){
                    t = tf.createFunctionTerm(aOP.getCastToInt(),t); 
                }
            }else if(PrimitiveType.JAVA_LONG.equals(type.getJavaType())){
                if(aOP.getCastToLong() != null){
                    t = tf.createFunctionTerm(aOP.getCastToLong(),t); 
                }
            }
        }
    |
        t=unaryexprnotplusminus
;

unaryexprnotplusminus returns [Term t=null]:
        "!" t=unaryexpr {t = tf.createJunctorTerm(Op.NOT,new Term[]{t});}
    |
        "~" t=unaryexpr 
        { 
            Function f = aOP.getNeg(isLong(t));
            t = tf.createFunctionTerm(f,new Term[]{t});
        }
    |
        t=postfixexpr
;

postfixexpr returns [Term t=null]
{
    tempkjt = getJavaInfo().getKeYJavaType(cld);
}:
        t=primaryexpr (t=primarysuffix[tempkjt, t])*
        {
            if(t == null){
                throw new NotDeclException("variable or method ",
                    packageName, getFilename(), getLine(), getColumn());
            }
        }
;

primaryexpr returns [Term t=null]
{
    KeYJavaType kjt = null;
    ProgramVariable v = null;
}:
        t=constant
    |   id:IDENT 
        {   
            tempkjt = getJavaInfo().getKeYJavaType(cld);
            if(LA(1)==LPAREN){
                methodname = id.getText();
                return null;
            }
            t=lookupVar(new Name(id.getText()));
            if(t == null || t.op() == null) {
                try{
                    kjt = getJavaInfo().getTypeByClassName(id.getText());
                }catch(RuntimeException e){}
                if(kjt == null){
                    packageName = id.getText();
                }else{
                    packageName = "";
                }
            }else{
                if(t.op() instanceof ProgramVariable){
                    kjt = ((ProgramVariable) t.op()).getKeYJavaType();
                    if(PrimitiveType.JAVA_BOOLEAN.equals(
                            ((ProgramVariable) t.op()).
                            getKeYJavaType().getJavaType())){
                        t=tf.createEqualityTerm(t,trueLitTerm);
                    }
                }else if(t.op() instanceof AttributeOp){
                    kjt=(((AttributeOp) t.op()).attribute()).getKeYJavaType();
                    if(PrimitiveType.JAVA_BOOLEAN.equals(kjt.getJavaType())){
                        t=tf.createEqualityTerm(t,trueLitTerm);
                    }
                }
            }
            tempkjt = kjt;
        }
    |   "true" { t = tf.createJunctorTerm(Op.TRUE); }
    |   "false" { t = tf.createJunctorTerm(Op.FALSE); }
    |   "null" { t = tf.createFunctionTerm(Op.NULL);} 
    |   t=jmlprimary 
    |   "this" 
        {
            if(staticMode){
                throw new KeYSemanticException("this in"+
                    " static context", getLine(), getColumn(), getFilename());
            }
            t = tf.createVariableTerm((ProgramVariable) prefix);
            tempkjt = ((ProgramVariable) prefix).getKeYJavaType();
        }
    |   t=new_expr
;

primarysuffix[KeYJavaType kjt, Term t] returns [Term result = null]{
    ListOfTerm l=SLListOfTerm.EMPTY_LIST;
    Term arg = null; 
    String mName = methodname;
    methodname = null;
}:
        (DOT id:IDENT
            {
                if(LA(1)==LPAREN){
                    methodname = id.getText();
                    packageName = "";
                    return t;
                }
                ProgramVariable v = null;
                if(kjt != null){
                    v = (ProgramVariable) getJavaInfo().
                            getAttribute(id.getText(), kjt);
                }
                if(v==null && kjt != null){
                    JMLClassSpec cs = 
                    services.getImplementation2SpecMap().getSpecForClass(kjt);
                    if(cs != null){
                        try{
                            v = cs.lookupModelField(new Name(id.getText()));
                        }catch(AmbigiousModelElementException e){
                            throw new KeYSemanticException(
                                "Ambigious Model Element name"
                                , getLine(), getColumn(), getFilename());
                        }
                    }
                }
                if ( v == null /*&& !("length").equals(id.getText())*/ ) {
                    KeYJavaType classType = null;
                    try{
                        classType = getJavaInfo().getTypeByClassName(
                            packageName + "." + id.getText());
                    }catch(RuntimeException e){}
                    if(classType == null){
                        packageName += "."+id.getText();
                    }else{
                        packageName = "";
                        this.tempkjt = classType;
                    }
                }else{
                    if(v.isStatic()){
                        result = tf.createVariableTerm(v);
                    }else{
                        result = tf.createAttributeTerm(v,t);
                    }
                    this.tempkjt = v.getKeYJavaType();
                    if(PrimitiveType.JAVA_BOOLEAN.equals(
                            this.tempkjt.getJavaType())){
                        result=tf.createEqualityTerm(result,trueLitTerm);
                    }
                }
            }
        )
    |
        "(" (l=expressionlist)? ")" 
        {
            ListOfKeYJavaType sig = getTypeList(l);
            ProgramMethod pm = getJavaInfo().getProgramMethod(kjt, mName, 
                sig, getJavaInfo().getKeYJavaType(cld));
            // maybe it's a model method
            if(pm == null && getJavaInfo().getKeYJavaType(cld) == kjt){
                JMLClassSpec cs = 
                    services.getImplementation2SpecMap().getSpecForClass(kjt);
                if(cs != null){
                    try{
                        pm = cs.lookupModelMethod(
                            new Name(kjt.getSort()+ "::" + mName));
                    }catch(AmbigiousModelElementException e){
                        System.err.println(e.toString());
                    }
                }
            }
            if(pm == null){
                throw new NotDeclException("method", 
                    mName, getFilename(), getLine(), getColumn());
            }   
            if(staticMode && !pm.isStatic()){
                throw new KeYSemanticException("Nonstatic method "+pm.name()+ 
                    " in static context", getLine(), getColumn(), 
                    getFilename());
            }
            Term[] sub;
            int i=0;
            if(pm.isStatic()){
                sub = new Term[sig.size()];
            }else{
                sub = new Term[sig.size()+1];
                if(t==null){
                    sub[i++] = tf.createVariableTerm((ProgramVariable) prefix);
                }else{
                    sub[i++] = t;
                }
            }
            IteratorOfTerm it = l.iterator();
            Term temp;
            while(it.hasNext()){
                temp = simplifyArgument(it.next());
                if(temp.sort() == Sort.FORMULA){
                    Term var = tf.createVariableTerm(new ProgramVariable(
                        new ProgramElementName("_param"+(paramVarCount++)),
                        getJavaInfo().getKeYJavaType(
                                PrimitiveType.JAVA_BOOLEAN)));
                    sub[i++] = 
                        createModelMethod("formula"+(mmCount++), var, 
                        null,tf.createEqualityTerm(temp, tf.createEqualityTerm(
                                var, trueLitTerm)));
                }else{
                    sub[i++] = temp;
                }
            }
            result = tf.createFunctionTerm(pm, sub);
            if(PrimitiveType.JAVA_BOOLEAN.equals(
                    pm.getKeYJavaType().getJavaType())){
                result = tf.createEqualityTerm(trueLitTerm, result);
            }
            this.tempkjt = pm.getKeYJavaType();
        }
    |
        "["arg = expression "]"
        {
            result = tf.createArrayTerm(ArrayOp.getArrayOp(t.sort()), t, arg);
            this.tempkjt = services.getTypeConverter().getKeYJavaType(result);
        }
;

new_expr returns [Term t=null]
{
    KeYJavaType kjt;
    ListOfTerm l;
    Term temp;
}:
        "new" kjt=type l=new_suffix
        {
            ListOfKeYJavaType sig = getTypeList(l);
            Constructor c = getJavaInfo().getKeYProgModelInfo().getConstructor(
                kjt, sig);
            if(!(c instanceof ConstructorDeclaration)){
                throw new NotSupportedExpressionException(
                    "Default Constructor");
            }
            ConstructorDeclaration cod = (ConstructorDeclaration) c;
            ProgramMethod cm = 
                new ProgramMethod(cod, kjt, getJavaInfo().getKeYJavaType(cld),
                                  PositionInfo.UNDEFINED);
            Term[] sub;
            int i=0;
            sub = new Term[sig.size()];
            IteratorOfTerm it = l.iterator();
            while(it.hasNext()){
                temp = simplifyArgument(it.next());
                if(temp.sort() == Sort.FORMULA){
                    Term var = tf.createVariableTerm(new ProgramVariable(
                        new ProgramElementName("_param"+(paramVarCount++)),
                        getJavaInfo().getKeYJavaType(
                                PrimitiveType.JAVA_BOOLEAN)));
                    sub[i++] = 
                        createModelMethod("formula"+(mmCount++), var, 
                        null,tf.createEqualityTerm(temp, tf.createEqualityTerm(
                                var, trueLitTerm)));
                }else{
                    sub[i++] = temp;
                }
            }
            t = tf.createFunctionTerm(cm, sub);
            if(PrimitiveType.JAVA_BOOLEAN.equals(
                    cm.getKeYJavaType().getJavaType())){
                t = tf.createEqualityTerm(trueLitTerm, t);
            }
            this.tempkjt = kjt;            
        }
    ;

new_suffix returns [ListOfTerm l = SLListOfTerm.EMPTY_LIST]:
        "(" ( l=expressionlist )? ")" 
    ;

expressionlist returns [ListOfTerm l = SLListOfTerm.EMPTY_LIST]
{
    Term t;
}:
        t=expression {l=l.append(t);} ("," t=expression {l=l.append(t);})* 
;

constant returns [Term t=null]:
        t=javaliteral
;

javaliteral returns [Term t=null]:
        t=integerliteral
;

integerliteral returns [Term t=null]:
        t=decimalintegerliteral
;

decimalintegerliteral returns [Term t=null]:
        t=decimalnumeral
;

decimalnumeral returns [Term t=null]:
        n:DIGITS {t = toZNotation(n.getText());}
;

jmlprimary returns [Term t=null]
{
    KeYJavaType dummykjt;
}:
        RESULT 
        {
            t=tf.createVariableTerm((ProgramVariable)result);
            if(PrimitiveType.JAVA_BOOLEAN.equals(
                        result.getKeYJavaType().getJavaType())){
                t=tf.createEqualityTerm(t,trueLitTerm);
            }
            tempkjt = result.getKeYJavaType();
        }
    |
        ("(" QUANTIFIER) => t=specquantifiedexpression
    |
        OLD "(" t=specexpression ")" 
        {
            t = getOld(t);
        }
    |
        NONNULLELEMENTS "(" t=specexpression ")"
        {
            t = nonNullElements(t);
        }
    |   INFORMAL_DESCRIPTION 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "informal description");
        }
    |   NOT_MODIFIED "(" storereflist ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\not_modified");
        }
    |   FRESH "(" spec_expression_list ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\fresh");
        }
    |   REACH "(" dummy = specexpression ")"
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\reach");
        }
    |   DURATION "(" dummy = expression ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\duration");
        }
    |   SPACE "(" dummy = specexpression ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\space");
        }
    |   WORKINGSPACE "(" dummy = expression ")"
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\working_space");
        }
//    |   QUANTIFIER "(" dummy = expression ")"
    |   TYPEOF "(" dummy = specexpression ")"
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\typeof");
        }
    |   ELEMTYPE "(" dummy = specexpression ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\elemtype");
        }
    |   TYPE "(" dummykjt = type ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\type");
        }
    |   LOCKSET
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\lockset");
        }
    |   IS_INITIALIZED "(" dummykjt=referencetype ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\is_initialized");
        }
    |   INVARIANT_FOR "(" dummy=specexpression ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\invariant_for");
        }
    |   ( "(" LBLNEG ) => "(" LBLNEG IDENT dummy=specexpression ")"
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\lblneg");
        }
    |   ( "(" LBLPOS ) => "(" LBLPOS IDENT dummy=specexpression ")" 
        {
            throw new NotSupportedExpressionException("JML construct "+
                "\\lblpos");
        }
    |   
        "(" t=expression ")"

;

specquantifiedexpression returns [Term t=null]
{
    Term a=null; 
    ListOfNamed decls=null;
}:
        "("
        q:QUANTIFIER decls=quantifiedvardecls {bindVars(decls);} ";" 
        (
            (predicate ";" specexpression) => a=predicate ";" t=specexpression
        |
            (";")? t=specexpression 
        )
        {
            t=buildQuantifierTerm(q.getText(), decls, a, t); 
            unbindVars(); 
        }
        ")"
;

quantifiedvardecls returns [ListOfNamed decls=SLListOfNamed.EMPTY_LIST]
{
    KeYJavaType kjt = null;
    ProgramVariable v = null;
}:
        kjt=typespec v=quantifiedvariabledeclarator[kjt] 
        { decls=decls.append(v); }
        ("," v=quantifiedvariabledeclarator[kjt] { decls=decls.append(v); })*
;

typespec returns [KeYJavaType kjt = null]{
    int dim = 0;
}:
        kjt=type 
        (dim = dims 
            {   
                if(dim != 0){
                    kjt = services.getJavaInfo().getKeYJavaType(
                        ArraySortImpl.getArraySortForDim(kjt.getSort(), dim,
                            services.getJavaInfo().getObjectSort(), 
                            services.getJavaInfo().getCloneableSort()));
                }
            }
        )?
;

dims returns [int dim = 0]:
        ("[" "]" {dim++;} )+
    ;

type returns [KeYJavaType kjt = null]:
        (builtintype) => kjt = builtintype
        |
        kjt = referencetype
;

referencetype returns [KeYJavaType kjt = null]
{
    String n = null;
}:
        n=name 
        {kjt = services.getJavaInfo().getKeYJavaType(n);
            if(kjt==null){
                kjt = services.getJavaInfo().getTypeByClassName(n);
                if(kjt==null){
                    throw new NotDeclException("type", 
                        n, getFilename(), getLine(), getColumn());
                }
            }
        }
;   

builtintype returns [KeYJavaType kjt = null]
{
    String type = null;    
}:
        (
            "byte" {type = "byte";}
        |
            "short" {type = "short";}
        |
            "int" {type = "int";}
        |
            "long" {type = "long";}
        |
            "boolean" {type = "boolean";}
        )
        {kjt = services.getJavaInfo().getKeYJavaType(type);
            if(kjt==null){
                throw new NotDeclException("type", 
                    type, getFilename(), getLine(), getColumn());
            }
        }
;

name returns [String s = null]:
        id:IDENT {s=id.getText();} (DOT id2:IDENT {s+="."+id2.getText();})*
;

quantifiedvariabledeclarator[KeYJavaType kjt] returns [ProgramVariable v=null]
{
    int dim = 0;
}:
        id:IDENT (dim=dims)?           
        {
            if(dim != 0){
                kjt = services.getJavaInfo().getKeYJavaType(
                    ArraySortImpl.getArraySortForDim(kjt.getSort(), dim, 
                        services.getJavaInfo().getObjectSort(),
                        services.getJavaInfo().getCloneableSort()));
            }
            v = new ProgramVariable(new ProgramElementName(id.getText()), kjt);
        }
;
  

