// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.java.abstraction.KeYJavaType;
    import de.uka.ilkd.key.java.*;
    import de.uka.ilkd.key.java.reference.*;
    import de.uka.ilkd.key.util.*;
    import java.util.HashSet;
    import de.uka.ilkd.key.util.ExtList;
    import de.uka.ilkd.key.java.declaration.ArrayDeclaration;
    import de.uka.ilkd.key.java.expression.literal.IntLiteral;

}

/** 
 * A parser for modifies clauses,
 * i.e. FieldReferences.
 */  

class ModifiesParser extends Parser;
options {
	importVocab=KeYLexer;
	k = 3;
	defaultErrorHandler=true;
}

{
	private Namespace ns;
	private HashSet fields = new HashSet();
	private JavaInfo javaInfo;
	private boolean isStatic = false;
	private String packagePrefix = new String();
    private KeYExceptionHandler keh;

	public ModifiesParser(TokenStream lexer,
			  Namespace ns,
        JavaInfo javaInfo,
        KeYExceptionHandler keh) {
	    this(lexer);
	    this.ns = ns;
	    this.javaInfo = javaInfo;
        this.keh = keh;
	}

	public ModifiesParser(TokenStream lexer,
			  Namespace ns,
        JavaInfo javaInfo) {
	    this(lexer, ns, javaInfo, new KeYRecoderExcHandler());
	}

    public void reportError(RecognitionException ex){
        keh.reportException(ex);
    }
		
	public Namespace namespace() {
	    return ns;
        }

	public JavaInfo getJavaInfo() {
	    return javaInfo;
	}

	public HashSet modifiesElements() {
	    return fields;
	}

}



top {ProgramElement f;} : f = fieldref {fields.add(f);} 
			   (COMMA f = fieldref {fields.add(f);})* ;

fieldref returns [ProgramElement p = null]  {String y = null;
					      KeYJavaType last = null;
					      ProgramElement arrIndex = null;
					      Integer z = null;
					     } :
		(
			((y = programvar_or_typeref)
			{ p = (ProgramVariable) ns.lookup(new Name(y));
			  if (p == null) {
                       		last = javaInfo.getKeYJavaType(y);
				if (last == null) {
					throw new 
			   		 RuntimeException("Program Variable or Typeref not found: "+y);
				}
				p = new TypeRef(last);
				isStatic = true;
       			  } else
			  last = ((ProgramVariable)p).getKeYJavaType();
			}
            		| LPAREN (id:IDENT DOT 
                    	{packagePrefix = packagePrefix+id.getText()+".";
                    	})* (y = programvar_or_typeref)
                	{
                    	  last = javaInfo.getKeYJavaType(packagePrefix+y);
			  if (last == null) {
				throw new 
			   		 RuntimeException("Program Variable or Typeref not found: "+y);
			  }
			  p = new TypeRef(last);
			  isStatic = true;
                	}
             		  RPAREN )
			((DOT y = programvar_or_typeref
			{
			  ProgramVariable x = javaInfo.getAttribute(y, last);
              if (x==null) { // for private attributes, though OCL has
			                 // no visibilities
		          x=javaInfo.getAttribute
		                (last.getSort().toString()+"::"+y, 
		                 last);
	          }
			  if (x == null) {
			    throw new
			    RuntimeException("Attribute "+y+" not found in "
						+last+"   "+p);
			  }
			  if (isStatic) {
				if (!x.isStatic()) {
					throw new
			    		RuntimeException("ProgramVariable "+x+" is not static but "+
							  "a static field was expected.");
				}
				isStatic = false;
			  }
			  p = new FieldReference(x, (ReferencePrefix) p);
			  last = ((ProgramVariable)x).getKeYJavaType();
			})
			| (LBRACKET (
			 (IDENT DOT)=>(

			   arrIndex = fieldref
			 {
			  ExtList l = new ExtList();
			  l.add(arrIndex);
			  p = new ArrayReference(l, (ReferencePrefix) p);
			  last = ((ArrayDeclaration) last.getJavaType()).getBaseType().getKeYJavaType();
			 }
			 ) 
			| 
                         (
			    indexId:IDENT
			 {
			  ExtList l = new ExtList();
			  ProgramVariable attribute = (ProgramVariable) ns.lookup(new Name(indexId.getText()));
			  IntLiteral i = null;
			  if (attribute != null) {
			  	l.add(attribute);
			  }
			  else {
				try {
				    new Integer(indexId.getText());
				}
				catch (NumberFormatException e) {
				    throw new RuntimeException("The index into this array is not an integer "+
							  "and not a recognizable element of the program.");
				}
				i = new IntLiteral(indexId.getText());
				l.add(i);
			  }
			  p = new ArrayReference(l, (ReferencePrefix) p);
			  last = ((ArrayDeclaration) last.getJavaType()).getBaseType().getKeYJavaType();
			 }) )

		         
			 RBRACKET)
			)*
		) ;

programvar_or_typeref returns [String pv = null]:
   id:IDENT 
	{pv = (String) id.getText(); 
	} ;

