// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.parser.jml.*;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.jml.*;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.collection.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.Term;

public class JMLPostAndInvPO extends JMLProofOblInputImpl{

    private Term po = null;
    private JavaModelMethod jmm = null;

    public JMLPostAndInvPO(JMLMethodSpec spec, String javaPath, boolean allInv){
	super(spec, javaPath, allInv);
	md = spec.getMethodDeclaration();
	jmm = new JavaModelMethod(spec.getProgramMethod(), javaPath);
    }

    public ProofList getPO(){
	return new ProofList(
	    new Proof(name(),
		      (Term) po,
		      createProblemHeader(),
		      initConfig.createTacletIndex(), 
		      initConfig.createBuiltInRuleIndex(),
                      initConfig.getServices()), name());
    }

    public void readProblem(ModStrategy mod) throws ProofInputException{
	initConfig.progVarNS().add(initConfig.getServices().
				   getImplementation2SpecMap().getProgVarNS());
	initConfig.progVarNS().add(UsefulTools.buildParamNamespace(md));
	po = ((JMLMethodSpec) spec).createDLFormula(true, allInv);
    }

    public boolean fitsToContract(Contract ct) {
	if (!(ct instanceof JMLMethodContract) || !allInv) {
	    return false;
	}
	return ((JMLMethodContract) ct).getSpec() == spec;
    }

    public boolean checkPurity(){
	return po != null ? 
	    (UsefulTools.purityCheck(po, initConfig.getServices().
				    getImplementation2SpecMap()) == null):
	    true;
    }

    public Object getObjectOfContract(){
	return allInv ? jmm : (Object) md;
    }

    public String name(){
	return "Post, Invariant "+(allInv ? "(including all invariants)" : 
				  "(only invariants for "+
				  ((JMLMethodSpec) spec).
				   getClassDeclaration().getName()+")")+
	    " and Constraint PO for spec:\n "+
	    spec.toString();
    }

}
