// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.parser.jml.*;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.jml.*;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.collection.*;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.logic.SetOfChoice;
import de.uka.ilkd.key.logic.SetAsListOfChoice;
import de.uka.ilkd.key.logic.IteratorOfNamed;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.util.KeYExceptionHandler;

import java.util.Iterator;

public abstract class JMLProofOblInputImpl implements JMLProofOblInput {

    protected JMLSpec spec = null;
    protected MethodDeclaration md = null;
    protected InitConfig initConfig = null;
    protected String javaPath = null;
    protected KeYExceptionHandler exceptionHandler = null;
    protected boolean allInv = false;

    protected JMLProofOblInputImpl(JMLSpec spec, String javaPath, 
				   boolean allInv){
	this.spec = spec;
	this.javaPath = javaPath;
	this.allInv = allInv;
    }

    /**
     * sets the exception handler used to collect occured errors
     */
    public void setExceptionHandler(KeYExceptionHandler keh){
	exceptionHandler = keh;
    }

    public boolean fitsToContract(Contract ct) {
	return false;
    }

    public boolean askUserForEnvironment(){
	return false;
    }

    /**
     * Checks if the methods used in the specification that underlies this
     * JMLProofOblInput are declared with the modifier pure.
     */
    public abstract boolean checkPurity();

    public void read(ModStrategy mod) throws ProofInputException{}

    /**
     * The Java Model is already initialized;
     */
    public String readJava() throws ProofInputException{
	return null;
    }

    /** returns the path to the Java model.
     */
    public String getJavaPath(){
	return javaPath;
    }

    /** set the initial configuration used to read an input. It may become
     * modified during reading depending on the modification strategy used
     * for reading.
     */
    public void setConfig(InitConfig i){
	initConfig = i;
    }

    /** Specs have already been read.
     */
    public void readSpecs(){
    }

    public SetOfChoice readActivatedChoices() throws ProofInputException{
	return SetAsListOfChoice.EMPTY_SET;
    }

    /** reads the include section and returns an Includes object.  
     */
    public Includes readIncludes() throws ProofInputException{
	return new Includes();
    }
    
    /** returns the name of the proof obligation input.
     */
    public String name(){
	return spec.toString();
    }

    public Object getObjectOfContract(){
	return md;
    }

    public void startProtocol() {
	// do nothing
    }

    public String createProgramVarsSection(){
	IteratorOfNamed it = initConfig.progVarNS().allElements().iterator();
	String result = "program variables {\n";
	while(it.hasNext()){
	    ProgramVariable pv = (ProgramVariable) it.next();
	    result += pv.sort()+" "+pv.name()+";\n";
	}
	result += "}\n";
	return result;
    }

    public String createJavaSection(){
	return "java {\""+getJavaPath()+"\"}\n";
    }

    public String createProblemHeader(){
	return createJavaSection()+"\n"+createProgramVarsSection()+"\n";
    }
}


