// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.proof.mgt.Contract;
import de.uka.ilkd.key.proof.mgt.ProofEnvironment;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.collection.SetOfString;
import de.uka.ilkd.key.gui.JMLSpecBrowser;
import java.util.HashMap;
import java.io.File;
import javax.swing.JOptionPane;
import de.uka.ilkd.key.util.ProgressMonitor;


public class JavaInput extends KeYUserProblemFile implements ProofOblInput{

    protected File path = null;

    public JavaInput(String name, File path, ProgressMonitor monitor) {
	super(name, path, monitor);
	this.path = path;
    }

    public ProofList getPO(){
	String javaPath = null;
	try{
	    javaPath = getJavaPath();
	}catch(ProofInputException e){
	    initConfig.getServices().getExceptionHandler().reportException(e);
	}
	JMLProofOblInput jmlPO = null;
	int option = JOptionPane.YES_OPTION;
	do{
	    JMLSpecBrowser sb = new JMLSpecBrowser(initConfig.getServices(),
						   javaPath);
	    jmlPO = sb.getPOI();
	    if(jmlPO == null){
		option = JOptionPane.showConfirmDialog
		    (null, "Do you really want to abort loading the proof?", 
		     "Abort Loading", 
		     JOptionPane.YES_NO_OPTION);
	    }
	}while(option == JOptionPane.NO_OPTION && jmlPO == null);
	if(jmlPO == null){
	    throw new RuntimeException("No Proof");
	}else{
	    jmlPO.setConfig(getInitConfig());
	}
	try{
	    jmlPO.readProblem(ModStrategy.MOD_ALL);
	}catch(ProofInputException e){
	    initConfig.getServices().getExceptionHandler().reportException(e);
	}
	ProofList pl = jmlPO.getPO();
	getInitConfig().getProofEnv().registerProof(jmlPO, pl);
	return pl;
    }

    public int getNumberOfChars() {
	return 0;
    }

    /** starts reading the input and modifies the InitConfig of this 
     * object with respect to the given modification
     * strategy. 
     */
    public void read(ModStrategy mod) throws ProofInputException{
	return;
    }

    public void readProblem(ModStrategy mod) throws ProofInputException{
	return;
    }

    public boolean askUserForEnvironment(){
	return false;
    }

    /** returns the path to the Java model.
     */
    public String getJavaPath() throws ProofInputException{
	return path.getAbsolutePath();
    }

    public SetOfChoice readActivatedChoices() throws ProofInputException{
	return SetAsListOfChoice.EMPTY_SET;
    }

    public Includes readIncludes() throws ProofInputException{
	RuleSource standard = RuleSource.initRuleFile("standardRules.key");
	Includes includes = new Includes();
	includes.put("standardRules", standard);
	return includes;
    }
    
    public Object getObjectOfContract(){
	return null;
    }

    public boolean fitsToContract(Contract ct){
	return false;
    }

    public void readFuncAndPred(ModStrategy mod) throws ProofInputException {
	return;
    }

    public void readSorts(ModStrategy mod) throws ProofInputException {
	return;
    }

    public void readRulesAndProblem(ModStrategy mod) throws ProofInputException{
	return;
    }

    public void readProof(ProblemLoader prl) throws ProofInputException {
	return;
    }
}
