// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.ProverFacade;
import de.uka.ilkd.key.proof.mgt.Contract;

public class OCLInvInhMethPO extends OCLProofOblInput {

    private ReprModelMethod currOpRepr;
    private String preText;
    private String postText;
    private String invParentText;
    private String invText;
    private String xmifile;

    private Term prePO  = null;
    private Term postPO = null;


    public OCLInvInhMethPO(ReprModelMethod aMethRepr,
			   String xmifile, 
			   String preText, String invText, 
			   String invParentText, String postText) {	
	super("BehavioralSubtyping "+aMethRepr.getName());
	this.currOpRepr=aMethRepr;
	this.xmifile=xmifile;
	this.invParentText=invParentText;
	this.preText=preText;
	this.postText=postText;
	this.invText=invText;
    }


    public void read(ModStrategy mod) {
	if (initConfig==null) {
	    throw new IllegalStateException("InitConfig not set.");
	}
	ProverFacade po2prover = initConfig.createFacadeForTransl();
	prePO = generateMethodPO(
	    po2prover, currOpRepr, "purepre", "(" 
	    + FunctionalityOnModel.normalizeConstraint(preText) + " and " 
	    + FunctionalityOnModel.normalizeConstraint(invText) + ") implies " 
	    + FunctionalityOnModel.normalizeConstraint(invParentText), "", 
	    null,
	    xmifile);
	postPO = generateMethodPO(
	    po2prover, currOpRepr, "prepostaspl", 
	    FunctionalityOnModel.normalizeConstraint(preText) + " and " 
	    + FunctionalityOnModel.normalizeConstraint(invText),  "(" 
	    + FunctionalityOnModel.normalizeConstraint(postText) + " and " 
	    + FunctionalityOnModel.normalizeConstraint(invParentText) 
	    + ") implies " + 
	    FunctionalityOnModel.normalizeConstraint(invText), null, xmifile);
    }


    public ProofList getPO() {
        if (prePO == null) throw new IllegalStateException("No prePO term");
        if (postPO == null) throw new IllegalStateException("No postPO term");
        String s = createProofHeader(currOpRepr.getContainingClass()
				     .getRootDirectory());
        return new ProofList(new Proof[]{
           new Proof("pre conditions "+currOpRepr.getName(), prePO, s,
                     initConfig.createTacletIndex(),
		     initConfig.createBuiltInRuleIndex(),
                     initConfig.getServices()),
           new Proof("post conditions "+currOpRepr.getName(), postPO, s,
                     initConfig.createTacletIndex(),
		     initConfig.createBuiltInRuleIndex(),
                     initConfig.getServices())
	}, name());
    }

    public ReprModelClass getModelClass() {
	return currOpRepr.getContainingClass();
    }

    public Object getObjectOfContract() {
	return getModelClass();
    }

    public boolean fitsToContract(Contract ct) {
	return false; //%%
    }

}
