// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.parser.jml.*;
import de.uka.ilkd.key.proof.init.ProofOblInput;
import de.uka.ilkd.key.proof.init.JMLPostAndInvPO;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.jml.*;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import de.uka.ilkd.key.casetool.ReprModelMethod;

public class JMLMethodContract implements MethodContract{

    private ReprModelMethod meth;    
    private JMLMethodSpec spec;
    private Modality modality;
    private String typeDescr;
    private Taclet lemma;
    private List proofs = new LinkedList();
    private ProofEnvironment env;

    //iff true all existing invariants are added to the pre- and postcondition
    //of the method contract, iff false only the invariants of the containing
    //type and the instance invariants (if meth is not static) for the 
    //corresponding object are added. History constraints are always added.
    public static final boolean allInvariants = true;

    public JMLMethodContract(ReprModelMethod meth, JMLMethodSpec spec, 
			     Modality modality, String typeDescr) {
	this.meth=meth;
	this.spec = spec;
	this.modality = modality;
	this.typeDescr=typeDescr;
    }


    public void setProofEnv(ProofEnvironment env) {
	if (this.env==null) {
	    this.env=env;
	} else if (this.env!=env) {
	    throw new IllegalStateException("Method Contract may belong to "
					    +"only one environment.");
	}
    }

    public ProofEnvironment getProofEnv() {
	return env;
    }

    public ReprModelMethod getModelMethod() {
	return meth;
    }

    public Taclet getLemmaAndRegister() {
	if (lemma==null && (modality==de.uka.ilkd.key.logic.op.Op.DIA ||
		modality==de.uka.ilkd.key.logic.op.Op.BOX)&&
	    !((JavaModelMethod) meth).isConstructor()) {
	    JMLMethodLemmaFactory ms 
		= new JMLMethodLemmaFactory(
		    env.getInitConfig(), spec, typeDescr, modality); 
//	    if(((JavaModelMethod) meth).isModel()){
//		lemma = ms.getTacletForModelMethod();
//	    }else{
		lemma = ms.getTaclet();
//	    }
	    if (lemma!=null) {
		env.registerRule(lemma, new RuleJustificationByMethodSpec(this));
	    }
	}
	return lemma;
    }

    public List getProofs() {
	return proofs;
    }

    public void addProofList(ProofList pl) {
	proofs.add(pl);
    }

    public String getPreText() {
	return "";
    }

    public String getPostText() {
	return "";
    }

    public Modality getModality() {
	return modality;
    }

    public Object getObjectOfContract() {
	return meth;
    }

    public ProofOblInput getProofOblInput() {
	return new JMLPostAndInvPO(spec, env.getJavaModel().getModelDir(),
				   allInvariants);
    }

    public void removeProofList(ProofList pl) {
	proofs.remove(pl);
    }

    public String toString() {
	return "[JML] context " + meth + (spec == null? "" :
					  "\npre: "+spec.getCompletePre(
					      true, true)+
					  "\npost: "+spec.getCompletePost(
					      true, true)+
					  "\nmodifies: " + spec.
					  replaceModelFieldsInAssignable() +
					  "\npi1: "+ ((JMLLemmaMethodSpec) 
					  spec).getPi1() +
					  "\nmodality: "+modality);
    }

    public JMLMethodSpec getSpec(){
	return spec;
    }

    public boolean equals(Object cmp) {
	if (!(cmp instanceof JMLMethodContract)) return false;
	MethodContract mc = (MethodContract)cmp;
	return meth.equals(getObjectOfContract()) && 
	    modality == mc.getModality(); 
    }

    public int hashCode() {
	int result = 0;
	result = 37 * result + meth.hashCode(); 
	return result;
    }

}
