// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.reuse;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.gui.KeYMediator;
import de.uka.ilkd.key.java.Services;
import java.util.List;


public class ReuseFindTaclet {

   private List reusePoints;
   private KeYMediator medi;
   private Services services;
   private Constraint userC;

   public ReuseFindTaclet(KeYMediator medi, List reusePoints) {
       this.medi = medi;
       this.reusePoints = reusePoints;
       services = medi.getServices();
       userC = medi.getUserConstraint().getConstraint();
   }


   public void applicableWhere(ReusePoint blank) {
      Goal g = blank.target();
      RuleApp templateApp = blank.getApp();

      IteratorOfConstrainedFormula formIter;

      if (templateApp.posInOccurrence().isInAntec()) 
         formIter = g.sequent().antecedent().iterator();
      else
         formIter = g.sequent().succedent().iterator();

      // but only if pos taclet
      while (formIter.hasNext()) {
         ConstrainedFormula cfma = formIter.next();
         PosInOccurrence start = 
            new PosInOccurrence(cfma, PosInTerm.TOP_LEVEL, g.sequent());
         recMatch(start, cfma.formula(), templateApp.rule(), blank);
      }
   }

   private void recMatch(PosInOccurrence pos, Term t, Rule rule,
                         ReusePoint blank) {
      
      TacletApp tentativeApp = isApplicable(rule, pos);
      if (tentativeApp != null) {
          checkIfInstsAndRecord(blank, pos, tentativeApp);
      }
      // Succ- and Antec- are only applicable at top-level of formula
      if (!(rule instanceof RewriteTaclet)) return;
      int i;
      for (i=0; i<t.arity(); i++) {
         recMatch(pos.down(i), t.sub(i), rule, blank);
      }
   }




   private void checkIfInstsAndRecord(ReusePoint blank, PosInOccurrence pos, 
                                      TacletApp tentativeApp) {

         ListOfTacletApp ifCandidates;
         if (!tentativeApp.ifInstsComplete()) {
             ifCandidates = tentativeApp.findIfFormulaInstantiations(
 	        blank.target().sequent(), services, userC);
         } else {
             ifCandidates = SLListOfTacletApp.EMPTY_LIST.prepend(tentativeApp);
         }
         
         IteratorOfTacletApp it = ifCandidates.iterator();
         while (it.hasNext()) {
             ReusePoint p = blank.initialize(pos, it.next(), medi);
             if ((p != null)){ // RuleApp can be transferred // && goodEnough!
                reusePoints.add(p);
             }
         }
   }


   private TacletApp isApplicable(Rule rule, PosInOccurrence pos) {
      if (rule instanceof FindTaclet) {
         //this is copied from TermTacletAppIndex :-/
	 Constraint c = pos.constrainedFormula().constraint();
	 if ( pos.termBelowMetavariable() != null ) {
	     c = c.unify( 
                pos.constrainedFormula().formula().subAt(pos.posInTerm()),
		pos.termBelowMetavariable());
	     if (!c.isSatisfiable()) return null;
	 }

         NoPosTacletApp app = NoPosTacletApp.createNoPosTacletApp(
            (FindTaclet)rule);
         app = app.matchFind(pos, c, services, userC);
         if (app==null) return null;
         
         return app.setPosInOccurrence(pos);
         
      } else 
         throw new RuntimeException("Not a FindTaclet: Re-Use Failed "+rule);

   }

}
