// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;


import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.SVSubstitute;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.reference.ThisReference;
import de.uka.ilkd.key.java.reference.SuperReference;
import de.uka.ilkd.key.util.Debug;


/**
 *  Ensures is not a 'this' reference
 */
public class ReferenceCondition extends VariableConditionAdapter {

    private SchemaVariable var;

    public ReferenceCondition(SchemaVariable var) {
	this.var = var;
	if (!var.isProgramSV()) {
	    throw new IllegalArgumentException("Illegal schema variable");
	}
    }

    /**
     * checks if the condition for a correct instantiation is fulfilled
     * @param var the template Variable to be instantiated
     * @param candidate the SVSubstitute which is a candidate for an
     * instantiation of var
     * @param svInst the SVInstantiations that are already known to be needed 
     * @return true iff condition is fulfilled
     */
    public boolean check(SchemaVariable var, 
			 SVSubstitute candidate, 
			 SVInstantiations svInst,
			 Services services) {
	return !(candidate instanceof ThisReference || 
		 candidate instanceof SuperReference);
    }

    public String toString () {
	return "schema variable " + var + " is non null ";
    }

}
