// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.UpdateOperator;
import de.uka.ilkd.key.logic.ListOfTerm;
import de.uka.ilkd.key.logic.SLListOfTerm;
import de.uka.ilkd.key.logic.IteratorOfTerm;
import de.uka.ilkd.key.util.Debug;

/** this class represents a pair of an update operator and the term
 * that is used to update the variable(attribute) of update operator
 */

public class UpdatePair {

    /** the update terms */
    private final Term update;
    
    public UpdatePair(Term update) {

	Debug.assertTrue(update.op() instanceof UpdateOperator, 
			 "updatepair:: simultaneous update expected");
	
	this.update = update; 
    }


    public int arity() {
	return update.arity() - 1;
    }
    
    public Term sub(int n) {
	if (n>=arity()) {
	    throw new IndexOutOfBoundsException();
	}
	return update.sub(n);
    }
    
    public UpdateOperator updateOperator() {
	return (UpdateOperator)update.op();
    }

    private boolean equalEntries(UpdatePair cmp) {
	if (cmp.update.op() != update.op()) {
	    return false;
	}
	final int arity = arity();
	for (int i = 0; i<arity; i++) {
	    if (!sub(i).equals(cmp.sub(i))) {
		return false;
 	    }
	}
	return true;
    }

    public boolean equals(Object o) {
	if (!(o instanceof UpdatePair)) {
	    return false;
	}
	UpdatePair cmp = (UpdatePair) o;	
	return equalEntries(cmp);
    }
    
    public int hashCode(){
    	int result = 17;
    	result = 37 * result;
    	return result;
    }

    public String toString() {
	StringBuffer result = new StringBuffer("{");
 	for (int i = 0; i<arity(); i+=2) {
  	    result.append(""+sub(i));
  	    result.append(":=");
  	    result.append(""+sub(i+1));

 	    if (i < arity() + 2) {  	
 		result.append(",");
 	    }
  	}
	result.append("}");
	return result.toString();
    }
}
