// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.rule.metaconstruct.arith;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.expression.literal.*;
import java.math.BigInteger;
import de.uka.ilkd.key.logic.op.AbstractMetaOperator;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.op.Op;
/** this class implements the interface for
 * MetaAdderators. MetaAdderators are used to do complex term
 * transformation when applying a taclet. Often these transformation
 * caanot be described with the taclet scheme (or trying to do so would
 * result in a huge number of rules)
 */
public class MetaInByte extends AbstractMetaOperator {

    private int arity;

    private TermFactory termFactory = TermFactory.DEFAULT;

    public MetaInByte() {
	super(new Name("#inByte"), 1);
    }


    /**
     * checks whether the top level structure of the given @link Term
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the @link Term is valid.
     */
    public boolean validTopLevel(Term term) {
	// a meta operator accepts almost everything
	return term.op() instanceof MetaInByte && term.arity()==arity();
    }


    /** calculates the resulting term. */
    public Term calculate(Term term, SVInstantiations svInst, Services services) {
	Term arg = term.sub(0);
	BigInteger bigIntArg=null;

	bigIntArg = new
	    BigInteger(convertToDecimalString(arg, services));

	BigInteger max = new
	    BigInteger(""+Byte.MAX_VALUE);

	BigInteger min = new
	    BigInteger(""+Byte.MIN_VALUE);

	boolean result1 = bigIntArg.compareTo(max) <= 0;
	boolean result2 = min.compareTo(bigIntArg) <= 0;

	if (result1 && result2)
	    return termFactory.createJunctorTerm(Op.TRUE);
	else
	    return termFactory.createJunctorTerm(Op.FALSE);

    }

}
