// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule.updatesimplifier;

import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.ArrayOp;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.rule.UpdateSimplifier;

/**
 * An abstract update simplification rule for application on array access
 * operators.
 * 
 * @author bubel
 */
public abstract class ApplyOnArrayAccess extends ApplyOnAccessTerm {

    /**
     * @param updateSimplifier
     */
    public ApplyOnArrayAccess(UpdateSimplifier updateSimplifier) {
        super(updateSimplifier);
    }

    /**
     * determines whether this rule is applicable for the pair of update and
     * target (the term on which the update will be applied) term
     * 
     * @param update
     *            the Update to be applied on the given term
     * @param target
     *            the Term on which the update is applied
     * @return true if the rule can be applied on the update, target pair
     */
    public Term apply(Update update, Term target) {

        logEnter(update, target);

        final ArrayOp targetOp = (ArrayOp) target.op();

        final PropagationResult pr = propagateUpdateToSubterms(update, target);
        final Term[] subs = pr.getSimplifiedSubterms();
        boolean changed = pr.hasChanged();        
        
        final TermFactory tf = UpdateSimplifierTermFactory.DEFAULT.
            getBasicTermFactory();       
        
        Term result = changed ? tf.createArrayTerm(targetOp, subs) : target;
        
        // returns the conditional star term and simplifies as far
        // as possible:
        ArrayOfAssignmentPair mayBeAliases = 
            update.getAssignmentPairs((Location)result.op());
           
        mayBeAliases = optimize(subs, mayBeAliases);
        
        result = createAliasResolutionTerm(subs, mayBeAliases, result);

        logExit(result);
        return result;
    }
}
