// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy.feature;

import de.uka.ilkd.key.logic.ConstrainedFormula;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.rule.IfFormulaInstSeq;
import de.uka.ilkd.key.rule.ListOfIfFormulaInstantiation;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.util.Debug;

/**
 * Binary feature that returns zero iff the if- and find-formula of a Taclet are
 * matched to different members of the sequent. This feature is only applicable
 * to taclets with exactly one if-formula.
 */
public class DiffFindAndIfFeature extends BinaryTacletAppFeature {
    
    /** the single instance of this feature */
    public static final Feature INSTANCE = new DiffFindAndIfFeature ();

    protected boolean filter ( TacletApp app, PosInOccurrence pos, Goal goal ) {
        Debug.assertFalse ( pos == null,
                            "Feature is only applicable to rules with find" );
        
        ListOfIfFormulaInstantiation list = app.ifFormulaInstantiations();
        
        Debug.assertFalse ( list == null );
        
        if ( list.size () != 1 ) {
            Debug.fail ( "Wrong number of if-formulas." );
            return true;
        }
        
        Debug.assertTrue ( list.head () instanceof IfFormulaInstSeq );
        
        IfFormulaInstSeq iffi = (IfFormulaInstSeq) list.head ();
        
        ConstrainedFormula findFormula = pos.constrainedFormula ();
        ConstrainedFormula ifFormula   = iffi.getConstrainedFormula();
        
        return pos.isInAntec () != iffi.inAntec ()
                || !findFormula.equals ( ifFormula );
    }

}
